import os 
import json
import multiprocessing as mp
import argparse
import csv

from . import Experiment
from . import EmpCpds
from . import example_parameters


def main():
    """
    This is the main function for the pipeline that implements the CLI using docopt

    Args:
        args (dict): the args generated from doctopt
    """    

    params = example_parameters.PARAMETERS

    parser = argparse.ArgumentParser(description='pcpfm, LC-MS end-to-end processing')
    parser.add_argument('subcommand', metavar='subcommand',
                        help='one of the subcommands: _____')
    parser.add_argument('-p', '--parameters')
    parser.add_argument('-m', '--mode', default=None)
    parser.add_argument('--ppm', default=5, type=int)
    parser.add_argument('-s', '--sequence')
    parser.add_argument('-c', '--cores', type=int)
    parser.add_argument('-MS2_dir')
    parser.add_argument('-f', '--filter')
    parser.add_argument('-j', '--project')
    parser.add_argument('-o', '--output')
    parser.add_argument('-i', '--input')
    parser.add_argument('--name_field', default='Name')
    parser.add_argument('--path_field', default='Filepath')
    parser.add_argument('--asari_command')
    parser.add_argument('-tm', '--table_moniker')
    parser.add_argument('-em', '--empCpd_moniker')
    parser.add_argument('-nm', '--new_moniker')
    parser.add_argument('-cb', '--color_by', default=[])
    parser.add_argument('-bb', '--by_batch')
    parser.add_argument('-mb', '--marker_by', default=[])
    parser.add_argument('-tb', '--text_by', default=[])
    parser.add_argument('--all')
    parser.add_argument('--pca')
    parser.add_argument('--tsne')
    parser.add_argument('--spearman'),
    parser.add_argument('--kendall'),
    parser.add_argument('--missing_feature_distribution')
    parser.add_argument('--missing_feature_percentiles')
    parser.add_argument('--median_correlation_outlier_detection')
    parser.add_argument('--missing_feature_outlier_detection')
    parser.add_argument('--intensity_analysis')
    parser.add_argument('--feature_distribution')
    parser.add_argument('--feature_outlier_detection')
    parser.add_argument('--interactive_plots', default=False)
    parser.add_argument('--save_plots', default=False)
    parser.add_argument('--khipu_isotopes')
    parser.add_argument('--khipu_charges')
    parser.add_argument('--khipu_extended_adducts')
    parser.add_argument('--khipu_adducts_pos')
    parser.add_argument('--khipu_adducts_neg')
    parser.add_argument('--khipu_rt_tolerance')
    parser.add_argument('--blank_value')
    parser.add_argument('--sample_value')
    parser.add_argument('--query_field')
    parser.add_argument('--blank_intensity_ratio')
    parser.add_argument('--drop_name')
    parser.add_argument('--drop_field')
    parser.add_argument('--drop_value')
    parser.add_argument('--drop_others')
    parser.add_argument('--qaqc_filter')
    parser.add_argument('--conversion_command')
    parser.add_argument('--preprocessing_config')
    parser.add_argument('--new_csv_path')
    parser.add_argument('--TIC_normalization_percentile')
    parser.add_argument('--normalize_value')
    parser.add_argument('--feature_retention_percentile')
    parser.add_argument('--interpolation_ratio')
    parser.add_argument('--ms2_dir')

    args = parser.parse_args()
    if args.parameters:
        params.update(
            json.load(open(args.parameters))
        )
    for k, v in args.__dict__.items():
        if v:
            params[k] = v
    params['multicores'] = min(mp.cpu_count(), params['multicores'])

    for k,v in params.items():
        if type(v) is str and v.endswith(".json"):
            params[k] = json.load(open(v))

    if args.subcommand != "preprocess" and args.subcommand != "assemble":
        if type(params['input']) is str:
            if not params['input'].endswith(".json"):
                params['input'] = os.path.join(os.path.abspath(params['input']), "experiment.json")


    if args.subcommand == "preprocess":
        preprocess_config = params['preprocessing_config']
        params["sequence"] = os.path.abspath(params["sequence"])
        with open(params['new_csv_path'], 'w+') as out_csv_fh:
            for x, entry in enumerate(csv.DictReader(open(params['sequence']))):
                for new_field, _d in preprocess_config["mappings"].items():
                    entry[new_field] = []
                    for new_value, _dd in _d.items():
                        found = False
                        for substring in _dd["substrings"]:
                            for field_to_search in _dd["search"]:
                                if substring in entry[field_to_search] and found is False:
                                    found = True
                                    entry[new_field].append(new_value)
                        if found is False:
                            if "else" in _dd:
                                entry[new_field].append(_dd["else"])
                    entry[new_field] = "_".join(entry[new_field])
                if os.path.exists(params["path_field"]):
                    pass
                else:
                    if os.path.exists(os.path.join(os.path.dirname(params['sequence']), entry[params["name_field"]] + ".mzML")):
                        entry["InferredPath"] = os.path.join(os.path.dirname(params['sequence']), entry[params["name_field"]] + ".mzML")
                    elif os.path.exists(os.path.join(os.path.dirname(params['sequence']), entry[params["name_field"]] + ".raw")):
                        entry["InferredPath"] = os.path.join(os.path.dirname(params['sequence']), entry[params["name_field"]] + ".raw")
                if x == 0:
                    writer = csv.DictWriter(out_csv_fh, fieldnames=entry.keys())
                    writer.writeheader()
                writer.writerow(entry)
    elif args.subcommand == "assemble":
        experiment = Experiment.Experiment.construct_experiment_from_CSV(
            os.path.join(os.path.abspath(params['output']), str(params['project'])),
            params['sequence'],
            params['mode'],
            filter=params['filter'],
            name_field=params['name_field'],
            path_field=params['path_field'],
            exp_config=params['experiment_config']
        )
    elif args.subcommand == "convert":
        experiment = Experiment.Experiment.load(params['input'])
        experiment.convert_raw_to_mzML(params['conversion_command'])
        experiment.save()
    elif args.subcommand == "asari":
        experiment = Experiment.Experiment.load(params['input'])
        experiment.asari(params['asari_command'])
        experiment.save()
    elif args.subcommand == "QAQC":
        experiment = Experiment.Experiment.load(params['input'])
        experiment.parameters = params["experiment_config"]
        feature_table = experiment.retrieve(params['table_moniker'], True, False, True)
        for cosmetic_param in ['color_by', 'text_by', 'marker_by']:
            if cosmetic_param in params:
                if type(params[cosmetic_param]) is str:
                    params[cosmetic_param] = json.loads(params[cosmetic_param])
                    if type(params[cosmetic_param]) is str:
                        params[cosmetic_param] = list(params[cosmetic_param])
            else:
                params[cosmetic_param] = []
        for display_option in ['interactive_plots', 'save_plots']:
            if display_option not in params:
                params[display_option] = False
        experiment.QCQA_results[params['table_moniker']] = feature_table.QAQC(params)
        experiment.save()
    elif args.subcommand == "summarize":
        experiment = Experiment.Experiment.load(params['input'])
        experiment.summarize()
    elif args.subcommand == "build_empCpds":
        experiment = Experiment.Experiment.load(params['input'])
        params['khipu_adducts'] = params['khipu_adducts_pos'] if experiment.ionization_mode == "pos" else params['khipu_adducts_neg']
        EmpCpds.empCpds.construct_empCpds_from_feature_table(experiment,
                                                             params['khipu_isotopes'],
                                                             params['khipu_adducts'],
                                                             params['khipu_extended_adducts'],
                                                             params['table_moniker'],
                                                             params['empCpd_moniker'],
                                                             False,
                                                             params['khipu_rt_tolerance'],
                                                             params['ppm'],
                                                             params['khipu_charges'])
        experiment.save()
    elif args.subcommand == "blank_masking":
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve(params['table_moniker'], True, False, True)
        feature_table.blank_mask(params['blank_value'],
                                 params['sample_value'],
                                 params['query_field'],
                                 params['filter'], 
                                 float(params['blank_intensity_ratio']), 
                                 params['by_batch'],
                                 params['batch_blanking_logic'])
        feature_table.save(params['new_moniker'])
    elif args.subcommand == "drop_samples":
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve(params['table_moniker'], True, False, True)
        if params['drop_name']:
            feature_table.drop_sample_by_name(params['drop_name'], params['drop_others'])
        elif params['filter']:
            feature_table.drop_samples_by_filter(params['filter'], params['drop_others'])
        elif params['qaqc_filter']:
            feature_table.drop_samples_by_qaqc(params['qaqc_filter'], params['drop_others'])
        elif params['drop_field'] and params['drop_value']:
            feature_table.drop_samples_by_field(params['drop_value'], params['drop_field'], params['drop_others'])
        feature_table.save(params['new_moniker'])
    elif args.subcommand == "normalize":
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve(params['table_moniker'], True, False, True)
        if params["TIC_normalization_percentile"]:
            feature_table.TIC_normalize(float(params["TIC_normalization_percentile"]),
                                        params["by_batch"],
                                        params["normalize_value"])
        feature_table.save(params['new_moniker'])
    elif args.subcommand == "drop_missing_features":
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve(params['table_moniker'], True, False, True)
        feature_table.drop_missing_features(params["by_batch"], 
                                            float(params["feature_retention_percentile"]),
                                            params["feature_drop_logic"])
        feature_table.save(params['new_moniker'])
    elif args.subcommand == "interpolate":
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve(params['table_moniker'], True, False, True)
        feature_table.interpolate_missing_features(float(params['interpolation_ratio']),
                                                   params['by_batch'],
                                                   params['interpolate_method'])
        feature_table.save(params['new_moniker'])
    elif args.subcommand == "batch_correct":
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve(params['table_moniker'], True, False, True)
        feature_table.batch_correct(params['by_batch'])
        feature_table.save(params['new_moniker'])
    elif args.subcommand == "delete":
        experiment = Experiment.Experiment.load(params['input'])
        if params["table_moniker"]:
            experiment.delete(args["table_moniker"], True, False)
        elif params['empCpd_moniker']:
            experiment.delete(args["empCpd_moniker"], True, False)
    elif args.subcommand == "log_transform":
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve(params['table_moniker'], True, False, True)
        feature_table.log_transform(params['log_transform_mode'])
        feature_table.save(params["new_moniker"])
    elif args.subcommand == "MS1_annotate":
        experiment = Experiment.Experiment.load(params['input'])
        if 'table_moniker' in params:
            if experiment.ionization_mode == "pos":
                adducts = params['feature_adducts_pos']
            else:
                adducts = params['feature_adducts_neg']
            feature_table = experiment.retrieve(params['table_moniker'], True, False, True)
            feature_table.MS1_annotate(
                params['targets'],
                float(params['annot_mz_tolerance']),
                float(params['annot_rt_tolerance']),
                params["search_isotopologues"],
                params["MS1_annotation_name"],
                adducts
            )
            feature_table.save(params['new_moniker'])
        if 'empCpd_moniker' in params:
            empCpd = experiment.retrieve(params['empCpd_moniker'], False, True, True)
            empCpd.MS1_annotate(params['targets'], 
                                float(params['annot_rt_tolerance']))
            empCpd.save(params['new_moniker'])
    elif args.subcommand == "MS2_annotate":
        experiment = Experiment.Experiment.load(params['input'])
        if experiment.ionization_mode == "pos":
            msp_file = params['msp_files_pos']
        elif experiment.ionization_mode == "neg":
            msp_file = params['msp_files_neg']
        if 'table_moniker' in params:
            feature_table = experiment.retrieve(params['table_moniker'], True, False, True)
            feature_table.MS2_annotate(
                msp_file,
                params['ms2_dir'],
                params["annot_mz_tolerance"],
                params["annot_rt_tolerance"],
                params["MS2_annotation_name"],
                params["ms2_similarity_metric"],
                params["ms2_min_peaks"],
                params["find_experiment_ms2"]
            )
            feature_table.save(params["new_moniker"])
        if 'empCpd_moniker' in params:
            empCpd = experiment.retrieve(params['empCpd_moniker'], False, True, True)
            empCpd.MS2_annotate(
                msp_file,
                params['ms2_dir'],
                params["annot_mz_tolerance"],
                params["annot_rt_tolerance"],
                params["ms2_similarity_metric"],
                params["ms2_min_peaks"],
                params["find_experiment_ms2"])
            empCpd.save(params["new_moniker"])
    elif args.subcommand == "retrieve":
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve(params['table_moniker'], True, False, False)
    elif args.subcommand == "log_transform":
        experiment = Experiment.Experiment.load(params['input'])
        feature_table = experiment.retrieve(params['table_moniker'], True, False, False) 
        feature_table.log_transform(params['new_moniker'], params["log_transform_mode"])


def CLI():
    #args = docopt(__doc__)
    main()

if __name__ == '__main__':
    #args = docopt(__doc__)
    main()

    


from khipu.epdsConstructor import epdsConstructor
from khipu.extended import isotope_search_patterns, extended_adducts, adduct_search_patterns, adduct_search_patterns_neg
from jms.dbStructures import ExperimentalEcpdDatabase, knownCompoundDatabase
from jms.io import read_table_to_peaks
from mass2chem.formula import PROTON
import matchms
import json
from intervaltree import IntervalTree
import os
from copy import deepcopy

class empCpds:
    def __init__(self, dict_empCpds, experiment, moniker):
        """
        the empCpds object is a wrapper around dict_empCpds that will associate the dict_empCpds with a moniker and 
        experiment object. 

        Args:
            dict_empCpds (dict): dictionary of empCpds
            experiment (Experiment): the experiment object for the experiment from which these empCpds were generated
            moniker (str): the moniker for this empCpd, used by experiment to store empCpds
        """        
        self.dict_empCpds = dict_empCpds
        self.experiment = experiment
        self.moniker = moniker
        
        self.feature_id_to_khipu_id = {}
        for kp_id, khipu in dict_empCpds.items():
            for peak in khipu["MS1_pseudo_Spectra"]:
                self.feature_id_to_khipu_id[peak['id_number']] = kp_id

        self.__mz_trees = {}
        self.__rt_trees = {}

    def get_mz_tree(self, mz_tolerance):
        if mz_tolerance not in self.__mz_trees:
            mz_tree = IntervalTree()
            for _, khipu in self.dict_empCpds.items():
                for peak in khipu["MS1_pseudo_Spectra"]:
                    mz_tree.addi(peak["mz"] - (peak["mz"]/1e6 * mz_tolerance), peak["mz"] + (peak["mz"]/1e6 * mz_tolerance), peak['id_number'])
            self.__mz_trees[mz_tolerance] = mz_tree
        return self.__mz_trees[mz_tolerance]

    def get_rt_tree(self, rt_tolerance):
        if rt_tolerance not in self.__rt_trees:
            rt_tree = IntervalTree()
            for _, khipu in self.dict_empCpds.items():
                for peak in khipu["MS1_pseudo_Spectra"]:
                    rt_tree.addi(peak["rtime"] - rt_tolerance, peak["rtime"] + rt_tolerance, peak['id_number'])
            self.__rt_trees[rt_tolerance] = rt_tree 
        return self.__rt_trees[rt_tolerance]

    def search_for_feature(self, query_mz=None, query_rt=None, mz_tolerance=None, rt_tolerance=None):
        """search_for_feature 

        Given a query_mz and query_rt with corresponding tolerances in ppm and absolute units respectively find all 
        features by id_number that have a matching mz and rtime. 

        _extended_summary_

        All search fields are optional but if none are provided then all the features will be considered matching. 
        The mz tolerance should be in ppm while the rtime tolerance should be provided in rtime units. 

        :param query_mz: the mz to search for, defaults to None
        :type query_mz: float, optional
        :param query_rt: the rtime to search for, defaults to None
        :type query_rt: float, optional
        :param mz_tolerance: the tolerance in ppm for the mz match, defaults to None
        :type mz_tolerance: float, optional
        :param rt_tolerance: the tolerance in absolute units for the rt match, defaults to None
        :type rt_tolerance: float, optional
        :return: list of matching feature IDs
        :rtype: list
        """        
        
        if query_mz and mz_tolerance:
            mz_matches = set([x.data for x in self.get_mz_tree(mz_tolerance).at(query_mz)])
            if query_rt is None or rt_tolerance is None:
                return mz_matches
        else:
            mz_matches = None
        if query_rt and rt_tolerance:
            rt_matches = set([x.data for x in self.get_rt_tree(rt_tolerance).at(query_rt)])
            if mz_matches is None or mz_tolerance is None:
                return rt_matches
        else:
            rt_matches = None
        return list(rt_matches.intersection(mz_matches))


    def save(self, save_as_moniker=None):
        """
        This method saves the empirical compound dictionary to the annotation_subdirectory. 
        The path is determined by the moniker for the empCpds object, however, an alternative moniker can be provided 
        which effectively saves a new empCpd object. This also updates the empCpds registry in the experiment with the 
        path to the stored json.

        Args:
            save_as_moniker (str, optional): an alternative moniker to which to save the table. Defaults to None. 
        """
        save_as_moniker = self.moniker if save_as_moniker is None else save_as_moniker
        self.experiment.empCpds[save_as_moniker] = os.path.join(self.experiment.annotation_subdirectory, save_as_moniker + "_empCpds.json")
        with open(self.experiment.empCpds[save_as_moniker], 'w+') as out_fh:
            json.dump(self.dict_empCpds, out_fh, indent=4)
        self.experiment.save()

    @staticmethod
    def load(moniker, experiment):
        """
        This method generates the empCpd object for the provided moniker.

        Args:
            experiment (Experiment object): the experiment for which the empCpd was generated
            moniker (string): the empCpd moniker to load

        Returns:
            empCpds: the empCpds object for the specified moniker
        """        
        return empCpds(json.load(open(experiment.empCpds[moniker])), experiment, moniker)

    @staticmethod
    def construct_empCpds_from_feature_table(experiment, isotopes=isotope_search_patterns, adducts=None, extended_adducts=extended_adducts, feature_table_moniker='full', empCpd_moniker='default', add_singletons=False, rt_search_window=2, mz_tolerance=5, charges=[1,2,3]):
        """_summary_

        Args:
            experiment (_type_): _description_
            feature_table_moniker (str, optional): _description_. Defaults to 'full'.
            empCpd_moniker (str, optional): _description_. Defaults to 'default'.
            add_singletons (bool, optional): _description_. Defaults to True.

        Returns:
            _type_: _description_
        """
        if experiment.ionization_mode == 'pos' and adducts is None:
            adducts = adduct_search_patterns
        elif experiment.ionization_mode == 'neg' and adducts is None:
            adducts = adduct_search_patterns_neg
        peaklist = read_table_to_peaks(experiment.feature_tables[feature_table_moniker], has_header=True, mz_col=1, rtime_col=2, feature_id=0)
        for p in peaklist:
            p['id'] = p['id_number']
            p['representative_intensity'] = None
        ECCON = epdsConstructor(peaklist, experiment.ionization_mode)
        dict_empCpds = ECCON.peaks_to_epdDict(
            isotopes,
            adduct_search_patterns,
            extended_adducts,
            mz_tolerance_ppm=mz_tolerance,
            rt_tolerance=rt_search_window,
            charges=charges
        )
        all_feature_ids = set()
        for empCpd in dict_empCpds.values():
            for peak in empCpd["MS1_pseudo_Spectra"]:
                all_feature_ids.add(peak['id_number'])
        if add_singletons:
            for peak in peaklist:
                if peak['id_number'] not in all_feature_ids:
                    peak['ion_relation'] = None
                    dict_empCpds[len(dict_empCpds)] = {'interim_id': len(dict_empCpds), 
                                                       'neutral_formula_mass': '', 
                                                       'neutral_formula': '', 
                                                       'MS1_pseudo_Spectra': [peak]}
        empCpd = empCpds(dict_empCpds, experiment, empCpd_moniker)
        empCpd.save()
        return empCpd

    def MS1_annotate(self, annotation_sources, rt_tolerance=5):
        """
        Given multiple annotation sources in the JSON format compliant with JMS, annotate based on neutral formula 
        match to the annotation sources.

        Args:
            annotation_sources (list[str]): list of filepaths to annotation sources in JSON format
            rt_tolerance (int, optional): the rt_toleance to be used by ExperimentalEcpdDatabase. Defaults to 5.
        """        
        EED = ExperimentalEcpdDatabase(mode=self.experiment.ionization_mode, rt_tolerance=rt_tolerance)
        EED.build_from_list_empCpds(list(self.dict_empCpds.values()))

        formula_entry_lookup = {}
        for source in annotation_sources:
            for entry in json.load(open(source)):
                if entry['neutral_formula'] not in formula_entry_lookup:
                    formula_entry_lookup[entry['neutral_formula']] = []
                formula_entry_lookup[entry['neutral_formula']].append(entry)
            KCD = knownCompoundDatabase()
            KCD.mass_index_list_compounds(json.load(open(source)))
            KCD.build_emp_cpds_index()
            EED.extend_empCpd_annotation(KCD)

        for empCpd in self.dict_empCpds.values():
            if 'mz_only_db_matches' not in empCpd:
                empCpd['mz_only_db_matches'] = []
            if 'list_matches' in empCpd and empCpd['list_matches']:
                for match in empCpd['list_matches']:
                    formula_mass, ion_relation, count = match
                    formula, mass = formula_mass.split("_")
                    if formula in formula_entry_lookup:
                        empCpd['mz_only_db_matches'].extend(formula_entry_lookup[formula])

    def MS2_annotate(self, msp_files, ms2_files, mz_tolerance=5, rt_tolerance=20, similarity_method='cosine_greedy', min_peaks=3, search_experiment=False):
        similarity_methods = {
            "cosine_greedy": matchms.similarity.CosineGreedy,
            "cosine_hungarian": matchms.similarity.CosineHungarian,
        }
        similarity_metric = similarity_methods[similarity_method]()
        observed_precursor_mzs = IntervalTree()
        observed_precursor_rts = IntervalTree()
        expMS2_registry = {}
        ms2_id = 0
        mzML = []
        if ms2_files:
            for directory, _ , files in os.walk(ms2_files):
                for file in files:
                    if file.endswith(".mzML"):
                        mzML.append(os.path.join(os.path.abspath(directory), file))
        if search_experiment:
            mzML += [x.mzml_filepath for x in self.experiment.acquisitions if 'dda' in x.mzml_filepath.lower()]

        for mzml_filepath in mzML:
            print("Extracting: ", mzml_filepath)
            try:
                for spectrum in matchms.importing.load_from_mzml(mzml_filepath, metadata_harmonization=False):
                    spectrum = matchms.filtering.add_precursor_mz(spectrum)
                    try:
                        precursor_mz = float(spectrum.get('precursor_mz'))
                        spectrum.set('retention_time', spectrum.metadata['scan_start_time'][0] * 60)
                        precursor_rt = float(spectrum.get('retention_time'))
                        spectrum = matchms.filtering.default_filters(spectrum)
                        spectrum = matchms.filtering.normalize_intensities(spectrum)
                        spectrum = matchms.filtering.add_precursor_mz(spectrum)
                        spectrum = matchms.filtering.require_minimum_number_of_peaks(spectrum, min_peaks)
                        ms2_id = len(expMS2_registry)
                        expMS2_registry[ms2_id] = {"spectrum": spectrum, 
                                                   "precursor_mz": precursor_mz, 
                                                   "precursor_rt": precursor_rt,
                                                    "origin": mzml_filepath, 
                                                    "Annotations": []}
                        observed_precursor_mzs.addi(precursor_mz - (precursor_mz / 1e6 * mz_tolerance * 2), precursor_mz + (precursor_mz / 1e6 * mz_tolerance * 2), ms2_id)
                        observed_precursor_rts.addi(precursor_rt - rt_tolerance, precursor_rt + rt_tolerance, ms2_id)
                    except:
                        pass
            except:
                pass
        print("Found: ", ms2_id, " spectra!")
        if ms2_id == 0:
            return
        hits = 0
        if type(msp_files) is str:
            msp_files = [msp_files]
        for msp_file in msp_files:
            for msp_spectrum in matchms.importing.load_from_msp(msp_file, metadata_harmonization=False):
                try:
                    precursor_mz = float(msp_spectrum.get('precursor_mz'))
                except:
                    precursor_mz = None
                if precursor_mz:
                    for expMS2_id in [x.data for x in observed_precursor_mzs.at(precursor_mz)]:
                        try:
                            msms_score, n_matches = similarity_metric.pair(expMS2_registry[expMS2_id]["spectrum"], msp_spectrum).tolist()
                            if msms_score > 0.60 and n_matches > min_peaks:
                                hits += 1
                                reference_id = msp_spectrum.get("compound_name")
                                if reference_id:
                                    expMS2_registry[expMS2_id]["Annotations"].append({
                                        "msms_score": msms_score,
                                        "matched_peaks": n_matches,
                                        "feature_id": None,
                                        "db_precursor_mz": precursor_mz,
                                        "origin": os.path.basename(msp_file),
                                        "reference_id": msp_spectrum.get("compound_name")
                                    })
                        except:
                            pass
        print("Found: ", hits, " hits")
        mapped = 0
        for expMS2_id, entry in expMS2_registry.items():
            if entry["Annotations"]:
                for feature_id in self.search_for_feature(entry["precursor_mz"], entry["precursor_rt"], 2 * mz_tolerance, rt_tolerance):
                    mapped += 1
                    khipu = self.dict_empCpds[self.feature_id_to_khipu_id[feature_id]]
                    entry_copy = deepcopy(entry)
                    entry_copy["Matching_Feature"] = feature_id
                    entry_copy["spectrum"] = [[x[0], x[1]] for x in spectrum.peaks]
                    if "MS2_Spectra" not in khipu:
                        khipu["MS2_Spectra"] = []
                    khipu["MS2_Spectra"].append(entry_copy)
        print("Mapped: ", mapped, " mapped to samples")

    def auth_std_annotate(self, auth_stds, mz_tolerance=5, rt_tolerance=30, rtime_permissive=False):
        """
        Given a list of authentic standards in JMS compliant format, annotate. 

        Args:
            auth_stds (list[str]): list of filepaths to auth standard JSON files
            mz_tolerance (int, optional): mz tolerance in ppm for matches. Defaults to 5.
            rt_tolerance (int, optional): rt tolerance, in absolute seconds, for matches. Defaults to 30.
            rtime_permissive (bool, optional): if true, rtime matching is not enforced. Defaults to False.
        """        
        mz_tree = IntervalTree()
        rt_tree = IntervalTree()

        for empCpd_id, empCpd in self.dict_empCpds.items():
            if empCpd['neutral_formula_mass'] == "":
                if self.experiment.ionization_mode == 'pos':
                    neutral_formula_mass = empCpd["MS1_pseudo_Spectra"][0]['mz'] - PROTON
                elif self.experiment.ioinzation_mode == 'neg':
                    neutral_formula_mass = empCpd["MS1_pseudo_Spectra"][0]['mz'] + PROTON
            else:
                neutral_formula_mass = empCpd["neutral_formula_mass"]
            mass_error = neutral_formula_mass / 1e6 * mz_tolerance
            mz_tree.addi(neutral_formula_mass - mass_error, neutral_formula_mass + mass_error, empCpd_id)
            rtimes = [peak['rtime'] for peak in empCpd["MS1_pseudo_Spectra"]]
            rt_tree.addi(min(rtimes) - rt_tolerance, max(rtimes) + rt_tolerance, empCpd_id)
        
        for auth_std in auth_stds:
            for std in json.load(open(auth_std)):
                if rtime_permissive:
                    rtime_empCpd_id_matches = set([x for x in self.dict_empCpds.keys()])
                else:
                    rtime_empCpd_id_matches = set()
                if 'retention_time' in std and std['retention_time']:
                    rtime_matches = rt_tree.at(std['retention_time'])
                    if rtime_matches:
                        rtime_empCpd_id_matches = set([x.data for x in rtime_matches])
                if rtime_empCpd_id_matches:
                    mz_empCpd_id_matches = set([x.data for x in mz_tree.at(std['neutral_formula_mass'])])
                    mz_rtime_empCpd_id_matches = mz_empCpd_id_matches.intersection(rtime_empCpd_id_matches)
                    for empCpd_id in mz_rtime_empCpd_id_matches:
                        if "identity" not in self.dict_empCpds[empCpd_id]:
                            self.dict_empCpds[empCpd_id]["identity"] = []
                        self.dict_empCpds[empCpd_id]["identity"].append(std)





import csv
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import scipy.stats
import os
import pandas as pd
import intervaltree 
from combat.pycombat import pycombat
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from matplotlib.patches import Patch
import matchms
import re
from mass2chem.formula import calculate_mass, PROTON, ELECTRON, parse_chemformula_dict
import csv
import json

class FeatureTable:
    def __init__(self, feature_table_filepath, experiment, moniker):
        """
        This object wraps a feature table

        Args:
            feature_table_filepath (str): path to the feature table on disk
            experiment (Experiment object): the experiment object for this feature table
        """
        self.feature_table_filepath = feature_table_filepath
        self.experiment = experiment
        self.moniker = moniker
        self.feature_table = pd.read_csv(
            feature_table_filepath, delimiter="\t")

        self.__mz_trees = {}
        self.__rt_trees = {}

    
    def get_mz_tree(self, mz_tol):
        if mz_tol not in self.__mz_trees:
            mz_tree = intervaltree.IntervalTree()
            for feature_id, mz in zip(self.feature_table["id_number"], self.feature_table["mz"]):
                mz_tree.addi(mz - mz / 1e6 * mz_tol, mz + mz / 1e6 * mz_tol, feature_id)
            self.__mz_trees[mz_tol] = mz_tree
        return self.__mz_trees[mz_tol]
    
    def get_rt_tree(self, rt_tol):
        if rt_tol not in self.__rt_trees:
            rt_tree = intervaltree.IntervalTree()
            for feature_id, rtime in zip(self.feature_table["id_number"], self.feature_table["rtime"]):
                rt_tree.addi(rtime - rt_tol, rtime + rt_tol, feature_id)
            self.__rt_trees[rt_tol] = rt_tree
        return self.__rt_trees[rt_tol]


    @property
    def sample_columns(self):
        """sample_columns 

        Return a list of the column names in the feature table that are sample names.

        _extended_summary_

        This is used when filtering the feature tables. When we search the experiment for a set of samples with a given
        filter, this returns samples in the experiment that may not be in the feature table. We can use this list to 
        filter out the samples in the experiment not in the feature table. 

        :return: sample_columns
        :rtype: list
        """        
        return [x for x in self.feature_table.columns if x in [a.name for a in self.experiment.acquisitions]]

    @property
    def non_sample_columns(self):
        """non_sample_columns 

        Return a list of the column names in the feature table that are sample names.

        _extended_summary_

        This is used when filtering the feature tables but typically the list of sample columns is used instead.

        :return: non_sample_columns
        :rtype: list
        """        
        return [x for x in self.feature_table.columns if x not in self.sample_columns]

    @property
    def log_transformed(self):
        """log_transformed 

        This property queries the experiment object to determine if the feature table has been log transformed already

        _extended_summary_

        Some operations log transform the feature table before analysis. Multiple log transforms would yield unwanted
        results so if an operation is going to log transform a feature table, check this first to ensure that it is 
        has not already been log transformed. 

        :return: is_log_transformed
        :rtype: boolean
        """        
        return self.moniker in self.experiment.log_transformed_feature_tables

    @staticmethod
    def load(moniker, experiment):
        """load 

        This method yields a FeatureTable object when given a feature table moniker. 

        _extended_summary_

        FeatureTables are registered with the experiment object using a moniker, a string that points to the file path
        for that feature table. This method queries the experiment object, gets the feature table path, and creates 
        the object. 

        :param moniker: the string with which the FeatureTable is registered
        :type moniker: string
        :param experiment: the experiment object with the FeatureTable
        :type experiment: object
        :return: the feature table for the moniker
        :rtype: FeatureTable
        """        
        return FeatureTable(pd.read_csv(open(experiment.feature_tables[moniker])), experiment, moniker)

    def make_nonnegative(self, fill_value=1):
        """make_nonnegative 

        This replaces all NaN and 0 values in the feature table with the specified fill_value

        _extended_summary_

        This is used primarially before log transforming the feature table to remove values that cannot be log transformed

        :param fill_value: the value to replace NaN and 0 with, defaults to 1
        :type fill_value: int, optional
        """        
        self.feature_table.fillna(0)
        for column in self.sample_columns:
            self.feature_table[column] = [x if x > 0 else fill_value for x in self.feature_table[column]]

    def save(self, new_moniker=None, drop_invariants=True):
        """save 

        Save the feature table as a pandas-created .tsv and register the new on-disk location with the experiment 
        object using the specified new_moniker or reuse the existing moniker. By default this drops features that 
        have no variance in the feature table. This can occur when a sample or samples are dropped and one or more
        features are zero or interpolated only in the remaining samples.

        _extended_summary_

        When an operation is performed that modifies a feature table, the resulting feature table can be saved to 
        disk using this method. The moniker for the feature table can be reused or a new moniker provided. If a new
        moniker is provided it cannot be preferred or full since we do not want to overwrite the asari results. 

        Dropping invariants is recommended to reduce the size of the feature table and prevent uninformative features
        from reaching downstream steps. There is no good reason to turn it off, but the option exists. 


        :param new_moniker: a new moniker to register the saved table with the experiment object, defaults to None
        :type new_moniker: string, optional
        :param drop_invariants: if true, drop features that have no variance, defaults to True
        :type drop_invariants: bool, optional
        """

        if new_moniker is None:
            new_moniker = self.moniker
            if new_moniker == "preferred" or new_moniker == "full":
                raise Exception("Cannot overwrite asari feature tables") 

        if drop_invariants:
            self.drop_invariants()
        try:
            output_path = os.path.join(
                self.experiment.filtered_feature_tables_subdirectory, new_moniker + "_Feature_table.tsv")
            self.feature_table.to_csv(os.path.join(
                self.experiment.filtered_feature_tables_subdirectory, output_path), sep="\t", index=False)
            self.experiment.feature_tables[new_moniker] = output_path
            self.experiment.save()
        except:
            print("FAILURE TO SAVE FEATURE TABLE")

    def save_fig_path(self, name):
        """save_fig_path 

        Given a desired name for a figure, this returns the path to which this figure should be saved.

        _extended_summary_

        This ensures that the resulting path for the figure is a reasonable path without special figures and is 
        saved to the appropriate location in the experiment directory. 

        :param name: desired name for the figure
        :type name: string
        :return: path to save figure
        :rtype: string
        """        
        fig_path = os.path.join(os.path.abspath(
            self.experiment.experiment_directory), "QAQC_figs/" + self.moniker + "/")
        if not os.path.exists(fig_path):
            os.makedirs(fig_path)
        name = re.sub(r"[/\\?%*:|\"<>\x7F\x00-\x1F]", "_", name)
        return os.path.join(fig_path, self.experiment.experiment_name + "_" + name + ".png")

    def gen_figure(self, figure_type, data, title='', x_label=None, y_label=None, params=None, skip_annot=False):
        """gen_figure 

        #todo - this needs to be cleaned up.

        A single method is used to generate the figures for the FeatureTable. This allows for consistent looking
        figures to be generated. 

        The permitted types of figures are:

        "bar" - make a bar plot
        "scatter" - make a scatter plot
        "clustermap" - make a clustermap using seaborn
        "heatmap" - make a heatmap

        

        _extended_summary_

        :param figure_type: _description_
        :type figure_type: _type_
        :param data: _description_
        :type data: _type_
        :param title: the title for the figure, defaults to ''
        :type title: str, optional
        :param x_label: string to apply to the x-axis, defaults to None
        :type x_label: str, optional
        :param y_label: string to apply to the y-axis, defaults to None
        :type y_label: str, optional
        :param params: _description_, defaults to None
        :type params: _type_, optional
        :param skip_annot: if true do not apply cosmetics to the figure, defaults to False
        :type skip_annot: bool, optional
        """        

        if params['interactive'] or params['save_figs']:
            colors = params['colors']
            markers = params['markers']
            text = params['text']
            if figure_type == "scatter":
                if type(data) is dict:
                    X = data.keys()
                    Y = data.values()
                else:
                    X = data[:, 0]
                    Y = data[:, 1]
                plt.title(title)
                plt.xlabel(x_label)
                plt.ylabel(y_label)
                if skip_annot is False:
                    if markers and colors:
                        for x, y, c, m in zip(X, Y, list(colors[0]), list(markers[0])):
                            plt.scatter(x, y, c=c, marker=m)
                    elif markers and not colors:
                        for x, y, m in zip(X, Y, list(colors[0])):
                            plt.scatter(x, y, marker=m)
                    elif colors and not markers:
                        for x, y, c in zip(X, Y, list(colors[0])):
                            plt.scatter(x, y, c=c)
                    else:
                        plt.scatter(X, Y)
                    if text:
                        for x, y, t in zip(X, Y, text[0]):
                            plt.text(x, y, t)
                else:
                    plt.scatter(X, Y)
                if params['marker_legend'] and skip_annot is False:
                    plt.tight_layout(rect=[0, 0, 0.75, 1])
                    handles = [mlines.Line2D(
                        [],
                        [],
                        color='k',
                        marker=v,
                        linestyle='None',
                        markersize=10,
                        label=k
                    ) for k, v in params['marker_legend'].items() if v in markers[0]]
                    handles += [Patch(facecolor=v, label=k)
                                for k, v in params['color_legend'].items() if v in colors[0]]
                    plt.legend(
                        handles=handles,
                        bbox_to_anchor=(1.0, 0.0),
                        bbox_transform=plt.gcf().transFigure,
                        loc='lower right'
                    )
            elif figure_type == "heatmap":
                if colors:
                    g = sns.clustermap(
                        data, col_colors=colors, yticklabels=y_label)
                else:
                    g = sns.clustermap(data, yticklabels=y_label)
                plt.suptitle(title)
                if params['color_legend']:
                    plt.tight_layout(rect=[0, 0, 0.75, 1])
                    plt.legend(
                        [Patch(facecolor=color)
                         for color in params['color_legend'].values()],
                        list(params['color_legend'].keys()),
                        bbox_to_anchor=(1.0, 0.0),
                        bbox_transform=plt.gcf().transFigure,
                        loc='lower right'
                    )
            elif figure_type == "clustermap":
                if colors:
                    sns.clustermap(data, col_colors=colors)
                else:
                    sns.clustermap(data)
                plt.suptitle(title)
                if params['color_legend']:
                    plt.tight_layout(rect=[0, 0, 0.75, 1])
                    plt.legend(
                        [Patch(facecolor=color)
                         for color in params['color_legend'].values()],
                        list(params['color_legend'].keys()),
                        bbox_to_anchor=(1.0, 0.0),
                        bbox_transform=plt.gcf().transFigure,
                        loc='lower right'
                    )
            elif figure_type == "bar":
                if False:
                    if text and colors:
                        plt.bar(
                            [x+"_"+str(i) for i, x in enumerate(text[0])], data[1], color=colors[0])
                    elif text and not colors:
                        plt.bar([x+"_"+str(i)
                                for i, x in enumerate(text[0])], data[1])
                    elif not text and colors:
                        plt.bar([i for i in range(len(data[1]))],
                                data[1], color=colors[0])
                    else:
                        plt.bar([i for i in range(data[1])], data[1])
                else:
                    if colors:
                        plt.scatter([i for i in range(len(data[1]))],
                                    data[1], color=colors[0])
                    else:
                        plt.scatter([i for i in range(len(data[1]))], data[1])
                plt.title(title)
                plt.xticks(rotation=90)
                plt.xlabel(x_label)
                plt.ylabel(y_label)
                if params['color_legend']:
                    plt.tight_layout(rect=[0, 0, 0.75, 1])
                    plt.legend(
                        [Patch(facecolor=color)
                         for color in params['color_legend'].values()],
                        list(params['color_legend'].keys()),
                        bbox_to_anchor=(1.0, 0.0),
                        bbox_transform=plt.gcf().transFigure,
                        loc='lower right'
                    )
            if params['save_figs']:
                plt.savefig(self.save_fig_path(title.replace(" ", "_")))
            if params['interactive']:
                plt.show()
            plt.clf()

    def search_for_feature(self, query_mz=None, query_rt=None, mz_tolerance=None, rt_tolerance=None):
        """search_for_feature 

        Given a query_mz and query_rt with corresponding tolerances in ppm and absolute units respectively find all 
        features by id_number that have a matching mz and rtime. 

        _extended_summary_

        All search fields are optional but if none are provided then all the features will be considered matching. 
        The mz tolerance should be in ppm while the rtime tolerance should be provided in rtime units. 

        :param query_mz: the mz to search for, defaults to None
        :type query_mz: float, optional
        :param query_rt: the rtime to search for, defaults to None
        :type query_rt: float, optional
        :param mz_tolerance: the tolerance in ppm for the mz match, defaults to None
        :type mz_tolerance: float, optional
        :param rt_tolerance: the tolerance in absolute units for the rt match, defaults to None
        :type rt_tolerance: float, optional
        :return: list of matching feature IDs
        :rtype: list
        """        
        
        if query_mz and mz_tolerance:
            mz_matches = set([x.data for x in self.get_mz_tree(mz_tolerance).at(query_mz)])
            if query_rt is None or rt_tolerance is None:
                return mz_matches
        else:
            mz_matches = None
        if query_rt and rt_tolerance:
            rt_matches = set([x.data for x in self.get_rt_tree(rt_tolerance).at(query_rt)])
            if mz_matches is None or mz_tolerance is None:
                return rt_matches
        else:
            rt_matches = None
        return list(rt_matches.intersection(mz_matches))

    def MS1_annotate(self, targets, mz_tolerance=5, rt_tolerance=20, search_isotopologues=False, annotation_column_name=None, adducts=None):
        adduct_list = []
        for adduct in adducts:
            mass = 0
            if "add" in adduct:
                mass += calculate_mass(parse_chemformula_dict(adduct["add"]))
            if "lose" in adduct:
                mass -= calculate_mass(parse_chemformula_dict(adduct["lose"]))
            for z in adduct["z"]:
                adduct_list.append([mass, adduct["name"], z])
        
        annotation_column_name = 'annotations' if annotation_column_name is None else annotation_column_name
        if annotation_column_name in self.feature_table.columns:
            annotations = {x: y for x, y in zip(self.feature_table['id_number'], self.feature_table[annotation_column_name])}
        else:
            annotations = {x: []for x in self.feature_table['id_number']}
        target_entries = []
        for target_file in targets:
            if target_file.endswith(".csv"):
                for entry in csv.DictReader(open(target_file)):
                    entry["Name"] = entry["\ufeffName"] if "\ufeffName" in entry else entry["Name"]
                    entry['Isotope Dictionary'] = {'[' + k + ']': v for k, v in json.loads(entry["Isotope Dictionary"]).items()}
                    entry['mass'] = calculate_mass(entry["Isotope Dictionary"])
                    target_entries.append(entry)
            elif target_file.endswith(".json"):
                for entry in json.load(open(target_file)):
                    entry["Name"] = entry["name"]
                    entry["Isotope Dictionary"] = parse_chemformula_dict(entry["neutral_formula"])
                    entry['mass'] = calculate_mass(entry["Isotope Dictionary"])
                    target_entries.append(entry)
        for entry in target_entries:
            num_12C = entry['Isotope Dictionary']['[12C]'] if '[12C]' in entry['Isotope Dictionary'] and search_isotopologues else 0
            if type(search_isotopologues) is int:
                num_12C = min(num_12C, search_isotopologues)
            for count_13C in range(num_12C + 1):
                C13_mass = calculate_mass({'[13C]': count_13C})
                for adduct in adduct_list:
                    iso_adduct_mass = (entry['mass'] + adduct[0] + C13_mass - (ELECTRON * adduct[2])) / abs(adduct[2])
                    if 'rtime' in entry:
                        matches = self.search_for_feature(iso_adduct_mass, entry['rtime'], mz_tolerance, rt_tolerance)
                    else:
                        matches = self.search_for_feature(iso_adduct_mass, None, 5, None)
                    for match in matches:
                        try:
                            isotopologue = 'M0' if count_13C == 0 else 'm+13C/12C*' + str(count_13C) + ",z=" + str(z)
                            annotations[match].append(entry['Name'] + "," + adduct[1] + "," + isotopologue)
                        except:
                            pass
        self.feature_table[annotation_column_name] = [annotations[x] for x in self.feature_table['id_number']]

    def MS2_annotate(self, msp_files, ms2_files, mz_tolerance=5, rt_tolerance=20, annotation_column_name="ms2_annotations", similarity_method='cosine_greedy', min_peaks=3, search_experiment=True):
        similarity_methods = {
            "cosine_greedy": matchms.similarity.CosineGreedy,
            "cosine_hungarian": matchms.similarity.CosineHungarian,
        }
        similarity_metric = similarity_methods[similarity_method]()
        observed_precursor_mzs = intervaltree.IntervalTree()
        observed_precursor_rts = intervaltree.IntervalTree()
        expMS2_registry = {}
        ms2_id = 0
        mzML = []
        if ms2_files:
            for directory, _ , files in os.walk(ms2_files):
                for file in files:
                    if file.endswith(".mzML"):
                        mzML.append(os.path.join(os.path.abspath(directory), file))
        if search_experiment:
            mzML += [x.mzml_filepath for x in self.experiment.acquisitions if 'dda' in x.mzml_filepath.lower()]

        for mzml_filepath in mzML:
            try:
                for spectrum in matchms.importing.load_from_mzml(mzml_filepath, metadata_harmonization=False):
                    spectrum = matchms.filtering.add_precursor_mz(spectrum)
                    try:
                        precursor_mz = float(spectrum.get('precursor_mz'))
                        spectrum.set('retention_time', spectrum.metadata['scan_start_time'][0] * 60)
                        precursor_rt = float(spectrum.get('retention_time'))
                        spectrum = matchms.filtering.default_filters(spectrum)
                        spectrum = matchms.filtering.normalize_intensities(spectrum)
                        spectrum = matchms.filtering.add_precursor_mz(spectrum)
                        spectrum = matchms.filtering.require_minimum_number_of_peaks(spectrum, min_peaks)
                        ms2_id = len(expMS2_registry)
                        expMS2_registry[ms2_id] = {"spectrum": spectrum, 
                                                   "precursor_mz": precursor_mz, 
                                                   "precursor_rt": precursor_rt, 
                                                   "origin": mzml_filepath, 
                                                   "Annotations": []}
                        observed_precursor_mzs.addi(precursor_mz - precursor_mz / 1e6 * mz_tolerance * 2, precursor_mz + precursor_mz / 1e6 * mz_tolerance * 2, ms2_id)
                        observed_precursor_rts.addi(precursor_rt - rt_tolerance, precursor_rt + rt_tolerance, ms2_id)
                    except:
                        pass
            except:
                pass
        if type(msp_files) is str:
            msp_files = [msp_files]
        for msp_file in msp_files:
            for msp_spectrum in matchms.importing.load_from_msp(msp_file, metadata_harmonization=False):
                try:
                    precursor_mz = float(msp_spectrum.get('precursor_mz'))
                except:
                    precursor_mz = None
                if precursor_mz:
                    for expMS2_id in [x.data for x in observed_precursor_mzs.at(precursor_mz)]:
                        try:
                            msms_score, n_matches = similarity_metric.pair(expMS2_registry[expMS2_id]["spectrum"], msp_spectrum).tolist()
                            if msms_score > 0.60 and n_matches > min_peaks:
                                reference_id = msp_spectrum.get("compound_name")
                                if reference_id:
                                    expMS2_registry[expMS2_id]["Annotations"].append({
                                        "msms_score": msms_score,
                                        "matched_peaks": n_matches,
                                        "feature_id": None,
                                        "db_precursor_mz": precursor_mz,
                                        "origin": msp_file,
                                        "reference_id": reference_id,
                                    })
                        except:
                            pass
        if annotation_column_name in self.feature_table.columns:
            annotations = {x: y for x,y in zip(self.feature_table['id_number'], self.feature_table[annotation_column_name])}
        else:
            annotations = {x: [] for x in self.feature_table['id_number']}

        for expMS2_id, entry in expMS2_registry.items():
            if entry["Annotations"]:
                for feature_id in self.search_for_feature(entry["precursor_mz"], entry['precursor_rt'], 2 * mz_tolerance, rt_tolerance):
                    for annot in entry["Annotations"]:
                        annotations[feature_id].append(annot["reference_id"] + ", score=" + str(msms_score)[:4])
        self.feature_table[annotation_column_name] = [';'.join(annotations[x]) for x in self.feature_table['id_number']]

    def median_correlation_outlier_detection(self, figure_params, correlation_type='pearson'):
        """
        The median correlation of a sample against all other samples can be expressed as a z-score against the median
        of ALL correlations in the experiment. A high or low Z-score indicates that the sample was poorly correlated 
        with other smaples in the experiment. 

        Args:
            feature_vector_matrix (np.ndarray): a selected feature matrix
            acquisition_names (list[str]): the names of the acquisitions co-indexed with the selected features matrix
            correlation_type (str, optional): determines the type of correlatio matric to calculate. Defaults to 'pearson'.
            interactive_plot (bool, optional): if interacitve, make plot in matlplotlib window. Defaults to False.

        Returns:
            result: dictionary storing the result of this QCQA operation
        """
        _fig_param = dict(figure_params)
        _fig_param["interactive"] = _fig_param["save_figs"] = False
        correlation_result = self.correlation_heatmap(
            _fig_param, correlation_type=correlation_type)
        all_correlations = []
        median_correlations = {}
        for sample_name_1, corr_dict in correlation_result["Result"].items():
            correlation_for_sample_name_1 = []
            for sample_name_2, corr_value in corr_dict.items():
                if sample_name_1 != sample_name_2:
                    correlation_for_sample_name_1.append(corr_value)
            median_correlations[sample_name_1] = np.median(
                correlation_for_sample_name_1)
            all_correlations.extend(correlation_for_sample_name_1)
        all_correlations_std = np.std(all_correlations)
        all_correlations_median = np.median(all_correlations)
        z_score_correlations = {name: (median_correlation - all_correlations_median) /
                                all_correlations_std for name, median_correlation in median_correlations.items()}

        self.gen_figure(
            "scatter",
            {i: v for i, v in enumerate(median_correlations.values())},
            title="Median Correlation Values for Samples",
            x_label="Sample",
            y_label="Median Correlation Value",
            params=figure_params
        )
        self.gen_figure(
            "scatter",
            {i: v for i, v in enumerate(z_score_correlations.values())},
            title="Median Correlation Z-Scores for Samples",
            x_label="Sample",
            y_label="Median Correlation Z-Score",
            params=figure_params
        )

        result = {
            "Type": "MedianCorrelationZScores",
            "Config": {},
            "Result": z_score_correlations
        }
        return result

    def intensity_analysis(self, figure_params):
        """
        Analyze mean, median, sum intensity values on the feature table

        Args:
            feature_vector_matrix (np.ndarray): a selected feature matrix
            acquisition_names (list[str]): the names of the acquisitions co-indexed with the selected features matrix
            interactive_plot (bool, optional): if interacitve, make plot in matlplotlib window. Defaults to False.

        Returns:
            result: dictionary storing the result of this QCQA operation
        """
        selected_ftable = self.feature_table[figure_params['acquisitions']].copy(
        )
        intensity_sums = np.sum(selected_ftable, axis=0)
        mean_feature_intensity = np.mean(selected_ftable, axis=0)
        median_feature_intensity = np.median(selected_ftable, axis=0)

        selected_ftable = selected_ftable.copy()
        selected_ftable[selected_ftable == 0] = np.nan
        filtered_mean_feature_intensity = np.nanmean(selected_ftable, axis=0)
        filtered_median_feature_intensity = np.nanmedian(
            selected_ftable, axis=0)

        log_selected_ftable = np.log2(selected_ftable)
        log_filtered_intensity_sum = np.nansum(log_selected_ftable, axis=0)
        log_filtered_mean_feature_intensity = np.nanmean(
            log_selected_ftable, axis=0)
        log_filtered_median_feature_intensity = np.nanmedian(
            log_selected_ftable, axis=0)

        TICs = np.nansum(selected_ftable, axis=0)
        log_TICs = np.log2(TICs)

        tables = [intensity_sums, mean_feature_intensity, median_feature_intensity, filtered_mean_feature_intensity, filtered_mean_feature_intensity,
                  log_filtered_intensity_sum, log_filtered_mean_feature_intensity, log_filtered_median_feature_intensity, log_TICs]
        titles = ["Sum Feature Intensity", "Mean Feature Intensity", "Median Feature Intensity",
                  "Mean Feature Intensity (dropped 0s)", "Median Feature Intensity (dropped 0s)",
                  "Log Sum Feature Intensities", "Log Mean Feature Intensity", "Log Median Feature Intensity",
                  "Log TICs"]
        paths = [x.lower().replace(" ", "_") for x in titles]

        for table, title, path in zip(tables, titles, paths):
            self.gen_figure(
                "bar",
                (figure_params['acquisitions'], table),
                title,
                x_label="title",
                y_label="sample",
                params=figure_params
            )

        result = {
            "Type": "IntensitySummary",
            "Config": {},
            "Result": {
                "SumIntensity": {name: value for name, value in zip(figure_params['acquisitions'], intensity_sums)},
                "MeanIntensity": {name: value for name, value in zip(figure_params['acquisitions'], mean_feature_intensity)},
                "MedianIntensity": {name: value for name, value in zip(figure_params['acquisitions'], median_feature_intensity)},
                "MissingDroppedSumIntensity": {name: value for name, value in zip(figure_params['acquisitions'], intensity_sums)},
                "MissingDroppedMeanIntensity": {name: value for name, value in zip(figure_params['acquisitions'], filtered_mean_feature_intensity)},
                "MissingDroppedMedianIntensity": {name: value for name, value in zip(figure_params['acquisitions'], filtered_median_feature_intensity)},
                "LogMissingDroppedSumIntensity": {name: value for name, value in zip(figure_params['acquisitions'], log_filtered_intensity_sum)},
                "LogMissingDroppedMeanIntensity": {name: value for name, value in zip(figure_params['acquisitions'], log_filtered_mean_feature_intensity)},
                "LogMissingDroppedMedianIntensity": {name: value for name, value in zip(figure_params['acquisitions'], log_filtered_median_feature_intensity)}
            }
        }
        return result

    def correlation_heatmap(self, figure_params, correlation_type, log_transform=True):
        """correlation_heatmap 

        Using a specified correlation function generate a correlation heatmap for the feature table. Optionally,
        log transform the feature table first. 

        _extended_summary_

        The permitted correlation types are: 

        "pearson", "spearman" or "kendall"

        Only pearson will log_transform the feature table if enabled since the non-parametric correlations will not 
        be affected by the log transform. 

        :param figure_params: dictionary with the figure params
        :type figure_params: dict
        :param correlation_type: what correlation type to use
        :type correlation_type: str
        :param log_transform: if true, log transform before linear correlation, defaults to True
        :type log_transform: bool, optional
        :return: a dictionary with the correlation results and configuration used to generate the result
        :rtype: dict
        """        
        correlation_mode = {
            "spearman": scipy.stats.spearmanr,
            "kendall": scipy.stats.kendalltau
        }

        feature_vector_matrix = self.feature_table[figure_params['acquisitions']].T
        if log_transform and not self.log_transformed:
            feature_vector_matrix = feature_vector_matrix + 1
            feature_vector_matrix = np.log2(feature_vector_matrix)
        if correlation_type == "pearson":
            corr_matrix = np.corrcoef(feature_vector_matrix)
        elif correlation_type == "spearman" or correlation_type == "kendall":
            corr_matrix = np.zeros(
                (feature_vector_matrix.shape[0], feature_vector_matrix.shape[0]))
            for i in range(feature_vector_matrix.shape[0]):
                for j in range(i, feature_vector_matrix.shape[0]):
                    corr = correlation_mode[correlation_type](feature_vector_matrix[i], feature_vector_matrix[j]).statistic
                    corr_matrix[i][j] = corr_matrix[j][i] = corr

        self.gen_figure(
            "clustermap",
            corr_matrix,
            title=correlation_type +
            "\n(log transformed)" if log_transform else correlation_type,
            x_label=figure_params['text'],
            y_label=figure_params['text'],
            params=figure_params
        )
        result = {
            "Type": "Correlation",
            "Config": {"Metric": correlation_type, "LogTransformed": log_transform},
            "Result": {figure_params['acquisitions'][i]: {figure_params['acquisitions'][j]: float(corr_matrix[i][j]) for j in range(corr_matrix.shape[0])} for i in range(corr_matrix.shape[0])}
        }
        return result

    def PCA(self, figure_params, log_transform=True):
        """
        Perform PCA on provided feature table

        Args:
            feature_vector_matrix (np.ndarray): a selected feature matrix
            acquisition_names (list[str]): the names of the acquisitions co-indexed with the selected features matrix
            interactive_plot (bool, optional): if interacitve, make plot in matlplotlib window. Defaults to False.

        Returns:
            result: dictionary storing the result of this QCQA operation
        """
        sample_ftable = self.feature_table[figure_params['acquisitions']].T
        scaler = StandardScaler()
        pca_embedder = PCA(n_components=2)
        if log_transform and not self.log_transformed:
            sample_ftable = np.log2(sample_ftable+1)
        pca_embedding = pca_embedder.fit_transform(
            scaler.fit_transform((sample_ftable)))
        self.gen_figure("scatter",
                        pca_embedding,
                        "PCA (n_components=2)",
                        x_label="PC 1 " +
                        str(round(pca_embedder.explained_variance_ratio_[
                            0] * 100, 1)) + "%",
                        y_label="PC 2 " +
                        str(round(pca_embedder.explained_variance_ratio_[
                            1] * 100, 1)) + "%",
                        params=figure_params)
        result = {
            "Type": "PCA",
            "Config": {"n_components": 2, "scaler": "StandardScaler"},
            "Result": {"Sample_Coord_Dict": {name: list(coord) for name, coord in zip(figure_params['acquisitions'], pca_embedding)}}
        }
        return result

    def TSNE(self, figure_params, perplexity=30):
        """
        Perform TSNE on provided feature table

        Args:
            feature_vector_matrix (np.ndarray): a selected feature matrix
            acquisition_names (list[str]): the names of the acquisitions co-indexed with the selected features matrix
            interactive_plot (bool, optional): if interacitve, make plot in matlplotlib window. Defaults to False.
            perplexity (int, optional): the perplexity of the TSNE plot, reduces by 1 until a solution is found. Defaults to 30.

        Returns:
            result: dictionary storing the result of this QCQA operation
        """
        try:
            tnse_embedded_vector_matrix = TSNE(n_components=2, perplexity=perplexity).fit_transform(
                self.feature_table[figure_params['acquisitions']].T)
            self.gen_figure(
                "scatter",
                tnse_embedded_vector_matrix,
                "TSNE (n_components=2)",
                x_label="Latent 1",
                y_label="Latent 2",
                params=figure_params
            )
            result = {
                "Type": "TSNE",
                "Config": {"n_components": 2},
                "Result": {"Sample_Coord_Dict": {name: [float(x) for x in coord] for name, coord in zip(figure_params['acquisitions'], tnse_embedded_vector_matrix)}}
            }
            return result
        except:
            if perplexity > 0:
                self.TSNE(figure_params, perplexity=perplexity-1)
            else:
                return {}

    def missing_feature_percentiles(self, figure_params):
        """
        Calculate the distribution of missing features with respect to percent of smaples with feature

        Args:
            feature_vector_matrix (np.ndarray): the selected feature matrix
            interactive_plot (bool, optional): if True, interactive plots are made. Defaults to False.

        Returns:
            result: dictionary storing the result of this QCQA operation
        """
        def __count_feature(row, columns):
            return np.sum([1 for x in row[columns] if x > 0])

        num_sample_with_feature = self.feature_table.apply(
            __count_feature, axis=1, args=(figure_params['acquisitions'],))
        percentile_table = []
        for percentile in range(101):
            num_samples_threshold = len(
                figure_params['acquisitions']) * percentile/100
            percentile_table.append([percentile, num_samples_threshold, int(
                np.sum(num_sample_with_feature <= num_samples_threshold))])
        self.gen_figure(
            "scatter",
            np.array([[x[0], x[2]] for x in percentile_table]),
            title="Missing Feature Percentiles",
            x_label="Percentile",
            y_label="Num. Dropped Features",
            params=figure_params,
            skip_annot=True
        )
        result = {
            "Type": "MissingFeaturePercentiles",
            "Config": {},
            "Result": {"PercentileTable": percentile_table}
        }
        return result

    def missing_feature_distribution(self, figure_params, intensity_cutoff=0):
        """
        Count the number of missing features or featuers below the specified intensity cutoff per features

        Args:
            feature_vector_matrix (np.ndarray): the selected feature matrix
            acquisition_names (list[str]): list of acquisition names
            intensity_cutoff (int, optional): values below this intesnity are considered missing. Defaults to 0.
            interactive_plot (bool, optional): if True, interactive plots are made. Defaults to False.

        Returns:
            result: dictionary storing the result of this QCQA operation
        """
        masked_ftables = self.feature_table[figure_params['acquisitions']
                                            ] <= intensity_cutoff
        missing_feature_counts = dict(zip(figure_params['acquisitions'], [
                                      0 for _ in figure_params['acquisitions']]))
        for name in figure_params['acquisitions']:
            for value in masked_ftables[name]:
                if value is True:
                    missing_feature_counts[name] += 1
        self.gen_figure(
            "bar",
            (figure_params['acquisitions'], [missing_feature_counts[name]
             for name in figure_params['acquisitions']]),
            title="Missing Feature Counts",
            x_label="Missing Feature Counts",
            y_label="Num. Missing Features",
            params=figure_params
        )
        result = {
            "Type": "MissingFeatureDistribution",
            "Config": {"intensity_cutoff": intensity_cutoff},
            "Result": {name: int(num_missing) for name, num_missing in missing_feature_counts.items()}
        }
        return result

    def feature_distribution(self, figure_params, intensity_cutoff=0):
        """
        Count the number of features above the specified intensity cutoff per features

        Args:
            feature_vector_matrix (np.ndarray): the selected feature matrix
            acquisition_names (list[str]): list of acquisition names
            intensity_cutoff (int, optional): values above this intensity are considered. Defaults to 0.
            interactive_plot (bool, optional): if True, interactive plots are made. Defaults to False.

        Returns:
            result: dictionary storing the result of this QCQA operation
        """
        masked_ftables = self.feature_table[figure_params['acquisitions']
                                            ] > intensity_cutoff
        feature_counts = dict(zip(figure_params['acquisitions'], [
                              0 for _ in figure_params['acquisitions']]))
        for name in figure_params['acquisitions']:
            for value in masked_ftables[name]:
                if value is True:
                    feature_counts[name] += 1
        self.gen_figure(
            "bar",
            (figure_params['acquisitions'], [feature_counts[name]
             for name in figure_params['acquisitions']]),
            title="Feature Counts",
            y_label="Num. Features",
            params=figure_params
        )
        result = {
            "Type": "FeatureDistribution",
            "Config": {"intensity_cutoff": intensity_cutoff},
            "Result": {name: int(num_missing) for name, num_missing in feature_counts.items()}
        }
        return result

    def feature_distribution_outlier_detection(self, figure_params, intensity_cutoff=0):
        """
        Count the number of features above the specified intensity cutoff per features and express as a Z-score based
        on feature count across all samples. 

        Args:
            feature_vector_matrix (np.ndarray): the selected feature matrix
            acquisition_names (list[str]): list of acquisition names
            intensity_cutoff (int, optional): values above this intensity are considered. Defaults to 0.
            interactive_plot (bool, optional): if True, interactive plots are made. Defaults to False.

        Returns:
            result: dictionary storing the result of this QCQA operation
        """
        _fig_param = dict(figure_params)
        _fig_param["interactive"] = _fig_param["save_figs"] = False
        feature_counts_result = self.feature_distribution(
            _fig_param, intensity_cutoff=intensity_cutoff)
        sample_names = [*feature_counts_result["Result"].keys()]
        feature_counts = np.array([*feature_counts_result["Result"].values()])
        feature_z_scores = (
            feature_counts - np.mean(feature_counts)) / np.std(feature_counts)
        self.gen_figure(
            "scatter",
            {i: z_score for i, z_score in enumerate(feature_z_scores)},
            title="Num Feature Z-Score",
            x_label="Sample",
            y_label="Num Feature Z-Score",
            params=figure_params
        )
        result = {
            "Type": "FeatureCountZScores",
            "Config": {"intensity_cutoff": intensity_cutoff},
            "Result": {name: float(z_score) for name, z_score in zip(figure_params['acquisitions'], feature_z_scores)}
        }
        return result

    def missing_feature_outlier_detection(self, figure_params, intensity_cutoff=0):
        """
        Count the number of features below the specified intensity cutoff per features and express as a Z-score based
        on missing feature count across all samples. 

        Args:
            feature_vector_matrix (np.ndarray): the selected feature matrix
            acquisition_names (list[str]): list of acquisition names
            intensity_cutoff (int, optional): values below this intensity are considered missing. Defaults to 0.
            interactive_plot (bool, optional): if True, interactive plots are made. Defaults to False.

        Returns:
            result: dictionary storing the result of this QCQA operation
        """
        missing_feature_counts_result = self.missing_feature_distribution(
            figure_params, intensity_cutoff=intensity_cutoff)
        # this relies upon the sorted order of the dictionary, may not be safe in all Python versions
        sample_names = [*missing_feature_counts_result["Result"].keys()]
        missing_feature_counts = np.array(
            [*missing_feature_counts_result["Result"].values()])
        missing_feature_z_scores = (
            missing_feature_counts - np.mean(missing_feature_counts)) / np.std(missing_feature_counts)
        self.gen_figure(
            "scatter",
            {i: z_score for i, z_score in enumerate(missing_feature_z_scores)},
            title="Num Missing Feature Z-Score",
            x_label="Sample",
            y_label="Num Missing Feature Z-Score",
            params=figure_params
        )
        result = {
            "Type": "MissingFeatureZScores",
            "Config": {"intensity_cutoff": intensity_cutoff},
            "Result": {name: float(z_score) for name, z_score in zip(sample_names, missing_feature_z_scores)}
        }
        return result

    def drop_invariants(self, zeros_only=False):
        """drop_invariants 

        This method drops features that have all zero intensity or the same intensity across all samples.

        _extended_summary_

        This situation occurs as a result of filtering. For instance if a contaiminant is only seen in the blanks, 
        when the blanks are dropped from the feature table, that feature is still in the table but will be zero (or an
        interpolated value) for the remaning samples. These features have no information and can complicate downstream
        analysis. 

        :param zeros_only: if true, only drop features that are all zero, defaults to False
        :type zeros_only: bool, optional
        """        
        def __filter_invariant(row, columns):
            values = set(list([x for x in row[columns]]))
            if len(values) == 1:
                if zeros_only and values[0] == 0:
                    return False
                else:
                    return False
            return True

        to_keep = []
        for keep_feature, id_number in zip(self.feature_table.apply(__filter_invariant, axis=1, args=(self.sample_columns,)), self.feature_table["id_number"]):
            if keep_feature:
                to_keep.append(id_number)
        self.feature_table = self.feature_table[self.feature_table['id_number'].isin(
            to_keep)].copy()

        for sample_column in self.sample_columns:
            values = set(list([x for x in self.feature_table[sample_column]]))
            if len(values) == 1:
                if zeros_only and values[0] == 0:
                    self.feature_table.drop(
                        columns=[sample_column], inplace=True)
                else:
                    self.feature_table.drop(
                        columns=[sample_column], inplace=True)

    def drop_sample_by_name(self, drop_name, drop_others=False):
        if drop_others:
            self.feature_table.drop(columns=[x for x in self.sample_columns if x != drop_name], inplace=True)
        else:
            self.feature_table.drop(columns=drop_name, inplace=True)

    def drop_samples_by_filter(self, filter, drop_others=False):
        to_drop = []
        for acquisition in self.experiment.filter_samples(filter):
            if acquisition.name in self.sample_columns:
                to_drop.append(acquisition.name)
        if drop_others:
            to_drop = [x for x in self.sample_columns if x not in to_drop]
        self.feature_table.drop(columns=to_drop, inplace=True)

    def drop_samples_by_field(self, value, field, drop_others=False):
        to_drop = [x.name for x in self.experiment.filter_samples({field: {"includes": [value]}})]
        if drop_others:
            to_drop = [x for x in self.sample_columns if x not in to_drop]
        self.feature_table.drop(columns=to_drop, inplace=True)

    def drop_samples_by_qaqc(self, qaqc_filter, drop_others=False):
        to_drop = []
        qaqc_results = {entry["Type"]: entry for entry in self.experiment.QCQA_results[self.moniker]}
        for field in qaqc_filter.keys():
            qaqc_results_for_field = qaqc_results[field]["Result"]
            if "GreaterThan" in qaqc_filter[field]["Conditions"]:
                max_value = qaqc_filter[field]["Conditions"]["GreaterThan"]
            if "LessThan" in qaqc_filter[field]["Conditions"]:
                min_value = qaqc_filter[field]["Conditions"]["LessThan"]
            for sample, value in qaqc_results_for_field.items():
                if min_value < value < max_value:
                    if qaqc_filter[field]["Action"] == "Keep":
                        to_drop.append(sample)
                else:
                    if qaqc_filter[field]["Action"] == "Drop":
                        to_drop.append(sample)
        if drop_others:
            to_drop = [x for x in self.sample_columns if x not in to_drop]
        if to_drop:
            self.feature_table.drop(columns=to_drop, inplace=True)

    def drop_samples(self, new_moniker=None, drop_types=[], type_field="Sample Type", drop_name=None, auto_drop=None):
        """drop_samples 

        This method drops samples from the feature table. This can be done in one of three ways.

        Option 1 is using a specified sample name. Option 2 is to drop a type of sample specified by a combination of 
        drop type and type field. Option 3 is based on a set of user-provided rules called auto_drop. 

        _extended_summary_

        Dropping a single sample name is straightforward. If the sample name is found, that column is dropped. 
        Dropping based on type is more complex. This approach checks the field specified by the type_field across all samples
        and the list of types in drop_types and if they match, that sample is dropped. 
        Finally, samples can be dropped based on their QA/QC results. Here a json file is provided to the CLI or 
        optionally a dictionary is provided via auto_drop directly that specifies allowed values of a given QAQC result
        by name using "GreaterThan" or "LessThan". More options will be added in the future. 

        :param new_moniker: new moniker to which to save the feature table, optional defaults to None
        :type new_moniker: str
        :param drop_types: list of types to drop based on type_Feild, defaults to []
        :type drop_types: list, optional
        :param type_field: the field to search for the types within, defaults to "Sample Type"
        :type type_field: str, optional
        :param drop_name: name of a sample to drop, defaults to None
        :type drop_name: str, optional
        :param auto_drop: dictionary with auto drop rules, defaults to None
        :type auto_drop: dict, optional
        """        
        if not auto_drop:
            drop = []
            if not drop_name:
                for drop_type in drop_types:
                    drop += list(self.experiment.filter_samples({type_field: {"includes": [drop_type]}}))
            elif drop_name:
                drop = [
                    a.name for a in self.experiment.acquisitions if a.name == drop_name]
        else:
            drop = []
            qaqc_results = {
                entry["Type"]: entry for entry in self.experiment.QCQA_results[self.moniker]}
            for field in auto_drop.keys():
                qaqc_results_for_field = qaqc_results[field]["Result"]
                if "GreaterThan" in auto_drop[field]["Conditions"]:
                    max_value = auto_drop[field]["Conditions"]["GreaterThan"]
                if "LessThan" in auto_drop[field]["Conditions"]:
                    min_value = auto_drop[field]["Conditions"]["LessThan"]
                for sample, value in qaqc_results_for_field.items():
                    if min_value < value < max_value:
                        if auto_drop[field]["Action"] == "Keep":
                            drop.append(sample)
                    else:
                        if auto_drop[field]["Action"] == "Drop":
                            drop.append(sample)
        if drop:
            self.feature_table.drop(columns=[x for x in drop if x in self.feature_table.columns], inplace=True)
        self.save(new_moniker if new_moniker is not None else self.moniker)

    def drop_others(self, new_moniker, keep_types, type_field="Sample Type", drop_name=None):
        """drop_others 

        Given a type of sample to keep by keep_types, drop all other fields

        # TO-DO this will be deprecated and merged with drop_samples

        _extended_summary_

        :param new_moniker: new moniker to which to save the feature table, optional defaults to None
        :type new_moniker: str
        :param drop_types: list of types to drop based on type_Feild, defaults to []
        :type drop_types: list, optional
        :param type_field: the field to search for the types within, defaults to "Sample Type"
        :type type_field: str, optional
        :param drop_name: name of a sample to drop, defaults to None
        :type drop_name: str, optional
        """        
        drop = []
        if not drop_name:
            for keep_type in keep_types:
                drop += list(self.experiment.filter_samples(
                    {type_field: {"lacks": [keep_type]}}))
        elif drop_name:
            drop = [
                a.name for a in self.experiment.acquisition if a.name != drop_name]
        else:
            pass
        drop = [x for x in drop if x in self.feature_table.columns]
        if drop:
            self.feature_table.drop(columns=drop, inplace=True)
        self.save(new_moniker)

    def blank_mask(self, 
                   blank_value="Blank",
                   sample_value="Unknown",
                   query_field="Sample Type",
                   filter=None,
                   blank_intensity_ratio=3,
                   by_batch=None,
                   logic_mode="or"):
        """blank_mask 

        Given a feature table containing samples that we consider blanks, drop all features in non-blank samples that
        do not have an intensity blank_intensity_ratio times higher than the mean intensity in the blanks.

        The blank samples are specified by the comibnation of blank_type and type_field. Non-blank samples are 
        specified by sample_type and type_field in a similar manner. 

        If there are batches in the experiment, blank masking is done per-batch. Then dropped if the ratio condition is
        not true in one sample (if logic_mode is "or") or in all samples if logic_mode is "and". The batches are 
        specified given a field in the metadata via the by_batch field.

        _extended_summary_

        :param new_moniker: _description_
        :type new_moniker: _type_
        :param by_batch: _description_, defaults to None
        :type by_batch: _type_, optional
        :param blank_intensity_ratio: _description_, defaults to 3
        :type blank_intensity_ratio: int, optional
        :param logic_mode: _description_, defaults to "or"
        :type logic_mode: str, optional
        :param blank_type: _description_, defaults to "Blank"
        :type blank_type: str, optional
        :param sample_type: _description_, defaults to "Unknown"
        :type sample_type: str, optional
        :param type_field: _description_, defaults to "Sample Type"
        :type type_field: str, optional
        """

        def __non_zero_mean(row, columns):
            non_zero_columns = [x for x in row[columns] if x > 0]
            return np.mean(non_zero_columns) if len(non_zero_columns) > 0 else 0

        def __any_logical(row, columns):
            return np.any(row[columns] == True)

        def __all_logical(row, columns):
            return np.all(row[columns] == True)

        blank_names = {x for x in self.experiment.filter_samples({query_field: {"includes": [blank_value]}}) if x in self.sample_columns}
        sample_names = {x for x in self.experiment.filter_samples({query_field: {"includes": [sample_value]}}) if x in self.sample_columns}
        blank_mask_columns = []
        if by_batch:
            for batch_name, batch_name_list in self.experiment.batches(by_batch).items():
                batch_blanks = [x for x in batch_name_list if x in blank_names]
                batch_samples = [
                    x for x in batch_name_list if x in sample_names]
                blank_means = self.feature_table.apply(
                    __non_zero_mean, axis=1, args=(batch_blanks,))
                sample_means = self.feature_table.apply(
                    __non_zero_mean, axis=1, args=(batch_samples,))
                to_filter = []
                for blank_mean, sample_mean in zip(blank_means, sample_means):
                    to_filter.append(
                        blank_mean * blank_intensity_ratio > sample_mean)
                blank_mask_column = "blank_masked_" + batch_name
                blank_mask_columns.append(blank_mask_column)
                self.feature_table[blank_mask_column] = to_filter
            if logic_mode == "and":
                self.feature_table["mask_feature"] = self.feature_table.apply(__all_logical, axis=1, args=(blank_mask_columns,))
            elif logic_mode == "or":
                self.feature_table["mask_feature"] = self.feature_table.apply(__any_logical, axis=1, args=(blank_mask_columns,))
            for blank_mask_column in blank_mask_columns:
                self.feature_table.drop(columns=blank_mask_column, inplace=True)
        else:
            blank_means = self.feature_table.apply(__non_zero_mean, axis=1, args=(list(blank_names),))
            sample_means = self.feature_table.apply(__non_zero_mean, axis=1, args=(list(sample_names),))
            to_filter = []
            for blank_mean, sample_mean in zip(blank_means, sample_means):
                to_filter.append(blank_mean * blank_intensity_ratio > sample_mean)
            blank_mask_column = "mask_feature"
            self.feature_table["mask_feature"] = to_filter
        self.feature_table = self.feature_table[self.feature_table["mask_feature"] == False]
        self.feature_table.drop(columns="mask_feature", inplace=True)


    def interpolate_missing_features(self, ratio=0.5, by_batch=None, method="min"):
        """interpolate_missing_features _summary_

        _extended_summary_

        :param new_moniker: _description_
        :type new_moniker: _type_
        :param ratio: _description_, defaults to 0.5
        :type ratio: float, optional
        :param by_batch: _description_, defaults to None
        :type by_batch: _type_, optional
        """        

        methods = {
            "min": min,
            "max": max,
        }

        def calc_interpolated_value(row, sample_names):
            values = [x for x in row[sample_names] if x > 0]
            if values:
                return methods[method](values) * ratio
            else:
                return 0

        sample_names = [
            a.name for a in self.experiment.acquisitions if a.name in self.feature_table.columns]
        if by_batch:
            for _, batch_name_list in self.experiment.batches(by_batch).items():
                filtered_batch_name_list = [
                    x for x in batch_name_list if x in sample_names]
                self.feature_table["feature_interpolate_value"] = self.feature_table.apply(calc_interpolated_value, axis=1, args=(filtered_batch_name_list,))
                for sample_name in filtered_batch_name_list:
                    self.feature_table[sample_name] = self.feature_table[[sample_name, "feature_interpolate_value"]].max(axis=1)
                self.feature_table.drop(columns="feature_interpolate_value", inplace=True)
        else:
            self.feature_table["feature_interpolate_value"] = self.feature_table.apply(calc_interpolated_value, axis=1, args=(sample_names,))
            for sample_name in sample_names:
                self.feature_table[sample_name] = self.feature_table[[sample_name, "feature_interpolate_value"]].max(axis=1)
            self.feature_table.drop(columns="feature_interpolate_value", inplace=True)

    def TIC_normalize(self, TIC_normalization_percentile=0.90, by_batch=None, normalize_mode='median'):
        """TIC_normalize _summary_

        _extended_summary_

        :param new_moniker: _description_
        :type new_moniker: _type_
        :param TIC_normalization_percentile: _descr iption_, defaults to 0.90
        :type TIC_normalization_percentile: float, optional
        :param by_batch: _description_, defaults to None
        :type by_batch: _type_, optional
        :param sample_type: _description_, defaults to "Unknown"
        :type sample_type: str, optional
        :param type_field: _description_, defaults to "Sample Type"
        :type type_field: str, optional
        :param normalize_mode: _description_, defaults to 'median'
        :type normalize_mode: str, optional
        """        
        function_map = {
            "median": np.median,
            "mean": np.mean,
        }
        if by_batch is not None:
            aggregate_batch_TICs = {}
            for batch_name, batch_name_list in self.experiment.batches(by_batch).items():
                batch_name_list = [
                    x for x in batch_name_list if x in self.feature_table.columns]
                self.feature_table["percent_inclusion"] = np.sum(
                    self.feature_table[batch_name_list] > 0, axis=1) / len(batch_name_list)
                TICs = {sample: np.sum(self.feature_table[self.feature_table["percent_inclusion"]
                                       > TIC_normalization_percentile][sample]) for sample in batch_name_list}
                norm_factors = {sample: function_map[normalize_mode](
                    list(TICs.values()))/value for sample, value in TICs.items()}
                aggregate_batch_TICs[batch_name] = function_map[normalize_mode](
                    list(TICs.values()))
                for sample, norm_factor in norm_factors.items():
                    self.feature_table[sample] = self.feature_table[sample] * norm_factor
            aggregate_batch_TIC_corrections = {batch: function_map[normalize_mode](list(
                aggregate_batch_TICs.values()))/value for batch, value in aggregate_batch_TICs.items()}
            for batch_name, batch_name_list in self.experiment.batches(by_batch).items():
                batch_name_list = [
                    x for x in batch_name_list if x in self.feature_table.columns]
                for sample in batch_name_list:
                    self.feature_table[sample] = self.feature_table[sample] * \
                        aggregate_batch_TIC_corrections[batch_name]
        else:
            sample_names = [x for x in self.feature_table.columns if x in [
                a.name for a in self.experiment.acquisitions]]
            self.feature_table["percent_inclusion"] = np.sum(
                self.feature_table[sample_names] > 0, axis=1) / len(sample_names)
            TICs = {sample: np.sum(self.feature_table[self.feature_table["percent_inclusion"]
                                   > TIC_normalization_percentile][sample]) for sample in sample_names}
            norm_factors = {sample: function_map[normalize_mode](
                list(TICs.values()))/value for sample, value in TICs.items()}
            for sample, norm_factor in norm_factors.items():
                self.feature_table[sample] = self.feature_table[sample] * norm_factor
        self.feature_table.drop(columns="percent_inclusion", inplace=True)

    def batch_correct(self, by_batch):
        """batch_correct _summary_

        _extended_summary_

        :param new_moniker: _description_
        :type new_moniker: _type_
        :param by_batch: _description_
        :type by_batch: _type_
        """        
        if len(self.experiment.batches(by_batch).keys()) > 1:
            batch_idx_map = {}
            for batch_idx, (_, acquisition_list) in enumerate(self.experiment.batches(by_batch).items()):
                for acquisition in acquisition_list:
                    batch_idx_map[acquisition] = batch_idx
            sample_names = [
                a.name for a in self.experiment.acquisitions if a.name in self.feature_table.columns]
            batches = [batch_idx_map[a.name]
                       for a in self.experiment.acquisitions if a.name in self.feature_table.columns]
            batch_corrected = pycombat(
                self.feature_table[sample_names], batches)
            for column in batch_corrected.columns:
                self.feature_table[column] = batch_corrected[column]
            self.make_nonnegative(fill_value=1)
        else:
            print("Unable to batch correct if only one batch!")
            raise Exception()

    def log_transform(self, new_moniker, log_mode="log2"):
        """log_transform _summary_

        _extended_summary_

        :param new_moniker: _description_
        :type new_moniker: _type_
        :param log_mode: _description_, defaults to "log2"
        :type log_mode: str, optional
        """        
        log_types = {
            "log10": np.log10,
            "log2": np.log2
        }
        try:
            self.experiment.log_transformed_feature_tables.append(new_moniker)
            self.experiment.save()
        except:
            self.experiment.log_transformed_feature_tables = [new_moniker]
            self.experiment.save()

        sample_names = [
            x.name for x in self.experiment.acquisitions if x.name in self.feature_table.columns]
        for sample_name in sample_names:
            self.feature_table[sample_name] = log_types[log_mode](
                self.feature_table[sample_name]+1)
        self.make_nonnegative()

    def drop_missing_features(self, by_batch=None, drop_percentile=0.8, logic_mode="or"):
        """drop_missing_features _summary_

        _extended_summary_

        :param new_moniker: _description_
        :type new_moniker: _type_
        :param by_batch: _description_, defaults to None
        :type by_batch: _type_, optional
        :param drop_percentile: _description_, defaults to 0.8
        :type drop_percentile: float, optional
        :param logic_mode: _description_, defaults to "or"
        :type logic_mode: str, optional
        :param sample_type: _description_, defaults to "Unknown"
        :type sample_type: str, optional
        :param type_field: _description_, defaults to "Sample Type"
        :type type_field: str, optional
        """        
        def __any(row, columns, drop_percentile):
            return not np.any(row[columns] >= drop_percentile)

        def __all(row, columns, drop_percentile):
            return not np.all(row[columns] >= drop_percentile)

        self.feature_table = self.feature_table.copy(deep=True)
        sample_names = [a.name for a in self.experiment.acquisitions]
        batch_columns = []
        if by_batch:
            for batch_name, batch_name_list in self.experiment.batches(by_batch).items():
                batch_column = "percent_inclusion_" + batch_name
                filtered_batch_name_list = [x for x in batch_name_list if x in self.feature_table.columns and x in sample_names]
                self.feature_table[batch_column] = np.sum(self.feature_table[filtered_batch_name_list] > 0, axis=1) / len(filtered_batch_name_list)
                batch_columns.append(batch_column)
            if logic_mode == "and":
                self.feature_table["drop_feature"] = self.feature_table.apply(__all, axis=1, args=(batch_columns, drop_percentile))
            elif logic_mode == "or":
                self.feature_table["drop_feature"] = self.feature_table.apply(__any, axis=1, args=(batch_columns, drop_percentile))
            #self.feature_table.drop(columns=batch_columns, inplace=True)
        else:
            sample_names = [x for x in sample_names if x in self.feature_table.columns]
            self.feature_table["drop_feature"] = (np.sum(self.feature_table[sample_names] > 0, axis=1) / len(sample_names)) < drop_percentile
        self.feature_table = self.feature_table[self.feature_table["drop_feature"] == False]
        self.feature_table.drop(columns="drop_feature", inplace=True)


    def generate_cosmetic(self, colorby=None, markerby=None, textby=None, seed=None):
        """generate_cosmetic _summary_

        _extended_summary_

        :param colorby: _description_, defaults to None
        :type colorby: _type_, optional
        :param markerby: _description_, defaults to None
        :type markerby: _type_, optional
        :param textby: _description_, defaults to None
        :type textby: _type_, optional
        :param seed: _description_, defaults to None
        :type seed: _type_, optional
        :return: _description_
        :rtype: _type_
        """        
        combined_cosmetic_map = {}
        if colorby:
            for cosmetic_map in [self.experiment.generate_cosmetic_map(c, 'color', seed) for c in colorby]:
                if cosmetic_map:
                    for k, v in cosmetic_map.items():
                        combined_cosmetic_map[("color", k)] = v
        if markerby:
            for cosmetic_map in [self.experiment.generate_cosmetic_map(m, 'marker', seed) for m in markerby]:
                if cosmetic_map:
                    for k, v in cosmetic_map.items():
                        combined_cosmetic_map[("marker", k)] = v
        colors = [[] for _ in colorby]
        markers = [[] for _ in markerby]
        texts = [[] for _ in textby]
        color_legend = {}
        marker_legend = {}
        for acquisition in self.experiment.acquisitions:
            if acquisition.name in self.feature_table.columns:
                for i, x in enumerate(colorby):
                    colors[i].append(combined_cosmetic_map[(
                        'color', acquisition.metadata_tags[x])])
                    color_legend[acquisition.metadata_tags[x]] = combined_cosmetic_map[(
                        'color', acquisition.metadata_tags[x])]
                for i, x in enumerate(markerby):
                    markers[i].append(combined_cosmetic_map[(
                        'marker', acquisition.metadata_tags[x])])
                    marker_legend[acquisition.metadata_tags[x]] = combined_cosmetic_map[(
                        'marker', acquisition.metadata_tags[x])]
                for i, x in enumerate(textby):
                    texts[i].append(acquisition.metadata_tags[x])
        return colors, markers, texts, color_legend, marker_legend

    def overlay_TICs(self):
        """overlay_TICs _summary_

        _extended_summary_
        """        
        TICs = []
        for acquisition in self.experiment.acquisitions:
            TIC = acquisition.TIC
            plt.scatter([x[0] for x in TIC], [x[1] for x in TIC])
            plt.show()
            print(acquisition.TIC)

    def QAQC(self, params):
        """
        This is the wrapper for all the qcqa functions. 

        Args:
            tag (str, optional): Sample type must match this field. Defaults to None.
            sort (bool, optional): if true, sort the sample names. Defaults to False.
            interactive (bool, optional): if True, interactive plots are generated. Defaults to False.

            These args control if the named step is performed or not in the curation:

            pca (bool, optional): Defaults to False.
            tsne (bool, optional): Defaults to False.
            pearson (bool, optional): Defaults to False.
            spearman (bool, optional): Defaults to False.
            kendall (bool, optional): Defaults to False.
            missing_feature_percentiles (bool, optional): Defaults to False.
            missing_feature_distribution (bool, optional): Defaults to False.
            median_correlation_outlier_detection (bool, optional): Defaults to False.
            missing_feature_outlier_detection (bool, optional): Defaults to False.
            intensity_analysis (bool, optional): Defaults to False.
            feature_distribution (bool, optional): Defaults to False.
            feature_outlier_detection (bool, optional): Defaults to False.

        Returns:
            list: with all qcqa results for the performed QCQA steps
        """

        colors, markers, texts, color_legend, marker_legend = self.generate_cosmetic(params['color_by'], params['marker_by'], params['text_by'])

        figure_params = {
            "acquisitions": list(self.sample_columns),
            "interactive": params['interactive_plots'],
            "save_figs": params['save_plots'],
            "text": texts,
            "markers": markers,
            "colors": colors,
            "color_legend": color_legend,
            "marker_legend": marker_legend
        }

        qcqa_result = []

        #if standards:
        #    qcqa_result.append(self.check_for_standards(
        #        figure_params, "/Users/mitchjo/Datasets/Standards/extraction_buffer_standards.csv"))
        #    qcqa_result.append(self.check_for_standards(
        #        figure_params, "/Users/mitchjo/Datasets/Standards/extraction_buffer_standards_with_lipidomix.csv"))
        if params['pca'] or params['all']:
            try:
                qcqa_result.append(self.PCA(figure_params))
            except:
                pass
        if params['tsne'] or params['all']:
            qcqa_result.append(self.TSNE(figure_params))
        if params['pearson'] or params['all']:
            try:
                qcqa_result.append(self.correlation_heatmap(
                    figure_params, correlation_type='pearson', log_transform=True))
            except:
                pass
        if params['kendall'] or params['all']:
            qcqa_result.append(self.correlation_heatmap(
                figure_params, correlation_type='kendall', log_transform=False))
        if params['spearman'] or params['all']:
            qcqa_result.append(self.correlation_heatmap(
                figure_params, correlation_type='spearman', log_transform=False))
        if params['missing_feature_percentiles'] or params['all']:
            qcqa_result.append(self.missing_feature_percentiles(figure_params))
        if params['missing_feature_distribution'] or params['all']:
            qcqa_result.append(self.missing_feature_distribution(figure_params))
        if params['missing_feature_outlier_detection'] or params['all']:
            qcqa_result.append(
                self.missing_feature_outlier_detection(figure_params))
        if params['median_correlation_outlier_detection'] or params['all']:
            qcqa_result.append(
                self.median_correlation_outlier_detection(figure_params))
        if params['intensity_analysis'] or params['all']:
            qcqa_result.append(self.intensity_analysis(figure_params))
        if params['feature_distribution'] or params['all']:
            qcqa_result.append(self.feature_distribution(figure_params))
        if params['feature_outlier_detection'] or params['all']:
            qcqa_result.append(self.feature_distribution_outlier_detection(figure_params))
        return qcqa_result



import csv
import shutil
import json
import os
import random
import matplotlib.colors as mcolors
import multiprocessing as mp
import pandas as pd
import random

from . import Acquisition
from . import FeatureTable
from . import EmpCpds


class Experiment:


    def __init__(self, 
                 experiment_name, 
                 experiment_directory,
                 acquisitions=None, 
                 skip_subdirectory_initialization=False, 
                 full_feature_table_path=None, 
                 preferred_feature_table_path=None,
                 qcqa_results=None,
                 drop_results=None,
                 ionization_mode=None,
                 organized_samples=None,
                 log=None,
                 feature_tables=None,
                 empCpds=None,
                 log_transformed_feature_tables=None,
                 cosmetics=None,
                 used_cosmetics=None,
                 extracted=None,
                 sub_dirs=None):
        """
        The Experiment object represents the set of acquisitions in an experiment and associated metadata.

        Args:
            experiment_name (str): a moniker for naming the experiment
            experiment_directory (str): path to the directory to store the experiment data and processing intermediates

            The following are used ONLY when loading an existing experiment

            acquisitions (list[Acquistions], optional): this is a list of Acquisition objects for the experiment. Defaults to None.
            skip_subdirectory_initialization (bool, optional): if True, do not create subdirectories for experiment. Defaults to False.
            full_feature_table_path (str, optional): path to full feature table #DEPRECATED - REMOVE. Defaults to None.
            preferred_feature_table_path (str, optional): path to preferred feature table #DEPRECATED - REMOVE. Defaults to None.
            qcqa_results (dict, optional): stores QCQA results per feature table. Defaults to None.
            drop_results (dict, optional): stores information needed to drop acquisitions during processing. Defaults to None.
            ionization_mode (str, optional): pos or neg, specifies the ionization mode of the MS for the experiment. Defaults to None.
            organized_samples (dict, optional): dictionary of acquisitions organized into types of samples #DEPRECATED - REMOVE. Defaults to None.
            log (logging_object, optional): logging instance, #DEPRECATED - REMOVE. Defaults to None.
            feature_tables (dict, optional): mapping of feature table monikers to feature tables on disk. Defaults to None.
            empCpds (dict, optional): mapping of empCpd monikers to empCpds on disk. Defaults to None.
        """        
        self.subdirectories =  {
            "asari_results": "asari_results/",
            "converted": "converted_acquisitions/",
            "raw": "raw_acquisitions/",
            "acquisition_data": "acquisitions/",
            "annotations": "annotations/",
            "filtered_feature_tables": "filtered_feature_tables/",
            "qaqc_figs": "QAQC_figs/"
        }

        # provided parameters
        self.experiment_name = experiment_name
        self.experiment_directory = os.path.abspath(experiment_directory)
        self.acquisitions = [] if acquisitions is None else acquisitions
        self.QCQA_results = {} if qcqa_results is None else qcqa_results
        self.organized_samples = {} if organized_samples is None else organized_samples
        self.drop_results = drop_results
        self.full_feature_table_path = full_feature_table_path
        self.preferred_feature_table_path = preferred_feature_table_path
        self.log = '' if log is None else log
        self.feature_tables = {} if feature_tables is None else feature_tables
        self.empCpds = {} if empCpds is None else empCpds
        self.log_transformed_feature_tables = [] if log_transformed_feature_tables is None else log_transformed_feature_tables
        self.cosmetics = {} if cosmetics is None else cosmetics
        self.used_cosmetics = [] if used_cosmetics is None else used_cosmetics
        self.extracted = [] if extracted is None else extracted
        self.__ionization_mode = ionization_mode
        if sub_dirs is not None:
            self.subdirectories = sub_dirs
        # generate subdirectories
        if not skip_subdirectory_initialization:
            if sub_dirs:
                self.subdirectories = sub_dirs
            self.initialize_subdirectories()

        # always computed
        self.asari_subdirectory = os.path.join(os.path.abspath(self.experiment_directory), self.subdirectories["asari_results"])
        self.converted_subdirectory = os.path.join(os.path.abspath(self.experiment_directory), self.subdirectories["converted"])
        self.raw_subdirectory = os.path.join(os.path.abspath(self.experiment_directory), self.subdirectories["raw"])
        self.annotation_subdirectory = os.path.join(os.path.abspath(self.experiment_directory), self.subdirectories["annotations"])
        self.filtered_feature_tables_subdirectory = os.path.join(os.path.abspath(self.experiment_directory), self.subdirectories["filtered_feature_tables"])

    @property
    def ionization_mode(self):
        if self.__ionization_mode is None:
            self.__ionization_mode = self.determine_ionization_mode()
        return self.__ionization_mode

    def initialize_subdirectories(self):
        """
        This function creates the subdirectories for the expriment. 

        The set of subdirectories to create is determiend by the class variable "subdirectories"
        """        
        to_create = [os.path.abspath(self.experiment_directory)]
        to_create += [os.path.join(to_create[0], subdir) for subdir in self.subdirectories.values()]
        for subdirectory_full_path in to_create:
            if not os.path.isdir(subdirectory_full_path):
                os.makedirs(subdirectory_full_path)

    def extract_all_acquisitions(self):
        workers = mp.Pool(mp.cpu_count())
        for data_path, acquisition in zip(workers.map(Acquisition.Acquisition.extract_acquisition, self.acquisitions), [x for x in self.acquisitions if x.name not in self.extracted]):
            if data_path is None:
                pass
            else:
                acquisition.data_path = data_path
                self.extracted.append(acquisition.name)
        self.save()

    @property
    def acquisition_datapath(self):
        return os.path.join(self.experiment_directory, self.subdirectories["acquisition_data"])

    @staticmethod
    def load(experiment_json_filepath):
        """
        Reconstitute the experiment object from a saved JSON file representing the object

        Args:
            experiment_json_filepath (str): path to the JSON file

        Returns:
            Experiment: an experiment object
        """        
        with open(experiment_json_filepath) as experiment_json_filehandle:
            JSON_repr = json.load(experiment_json_filehandle)
            acquisitions = []
            for acquisition_JSON in JSON_repr["acquisitions"]:
                acquisition = Acquisition.Acquisition(acquisition_JSON["name"], 
                                         acquisition_JSON["source_filepath"], 
                                         acquisition_JSON["metadata"])
                acquisition.mzml_filepath = acquisition_JSON["mzml_filepath"]
                acquisition.raw_filepath = acquisition_JSON["raw_filepath"]
                acquisition.spectra = acquisition_JSON["spectra"]
                acquisition.__min_mz = acquisition_JSON["__min_mz"]
                acquisition.__max_mz = acquisition_JSON["__max_mz"]
                acquisition.__min_rt = acquisition_JSON["__min_rt"]
                acquisition.__max_rt = acquisition_JSON["__max_rt"]
                acquisition.__TIC = acquisition_JSON["__TIC"] if "__TIC" in acquisition_JSON else None
                acquisition.data_path = acquisition_JSON["data_path"] if "data_path" in acquisition_JSON else None
                acquisitions.append(acquisition)
            #JSON_repr = defaultdict(JSON_repr, None)
            experiment = Experiment(JSON_repr["experiment_name"], 
                                    JSON_repr["experiment_directory"], 
                                    acquisitions=acquisitions, 
                                    skip_subdirectory_initialization=True,
                                    full_feature_table_path=JSON_repr["full_feature_table_path"],
                                    preferred_feature_table_path=JSON_repr["preferred_feature_table_path"],
                                    qcqa_results=JSON_repr["qcqa_results"],
                                    drop_results=JSON_repr["drop_results"],
                                    ionization_mode=JSON_repr["__ionization_mode"],
                                    organized_samples=JSON_repr["organized_samples"],
                                    feature_tables=JSON_repr["feature_tables"] if "feature_tables" in JSON_repr else None,
                                    empCpds=JSON_repr["empCpds"] if "empCpds" in JSON_repr else None,
                                    log_transformed_feature_tables=JSON_repr["log_transformed_feature_tables"] if "log_transformed_feature_tables" in JSON_repr else None,
                                    cosmetics=JSON_repr["cosmetics"] if "cosmetics" in JSON_repr else None,
                                    used_cosmetics=JSON_repr["used_cosmetics"] if "used_cosmetics" in JSON_repr else None,
                                    extracted=None,
                                    sub_dirs=JSON_repr['subdirectories'] #JSON_repr["extracted"] if "extracted" in JSON_repr else None
                                    )
            for acquisition in experiment.acquisitions:
                acquisition.experiment = experiment
        return experiment
        
    def delete(self, moniker, delete_feature_table=False, delete_empCpds=False):
        """
        Given a moniker for a feature table or empCpds, delete the entry on disk and in the object

        Args:
            moniker (str): the moniker to be deleted
            delete_feature_table (bool, optional): if True, delete the corresponding feature table. Defaults to False.
            delete_empCpds (bool, optional): if True, delete the corresponding empCpd. Defaults to False.
        """        
        if delete_feature_table:
            os.remove(self.feature_tables[moniker])
            del self.feature_tables[moniker]
        elif delete_empCpds:
            os.remove(self.empCpds[moniker])
            del self.empCpds[moniker]

    def retrieve(self, moniker, feature_table=False, empCpds=False, as_object=False):
        if feature_table:
            feature_table_path = self.feature_tables[moniker]
            if as_object:
                return FeatureTable.FeatureTable(feature_table_path, self, moniker)
            else:
                return feature_table_path
        elif empCpds:
            empCpd_path = self.empCpds[moniker]
            if as_object:
                return EmpCpds.empCpds.load(moniker, self)
            else:
                return empCpd_path

    def save(self):
        """
        Serialize the experiment and acquisitions and store it on disk
        """        
        save_path = os.path.join(self.experiment_directory, "experiment.json.tmp")
        with open(os.path.join(self.experiment_directory, "experiment.json.tmp"), "w") as save_filehandle:
            JSON_repr = {
                "experiment_directory": self.experiment_directory,
                "experiment_name": self.experiment_name,
                "acquisitions": [],
                "full_feature_table_path": self.full_feature_table_path,
                "preferred_feature_table_path": self.preferred_feature_table_path,
                "qcqa_results": self.QCQA_results,
                "drop_results": self.drop_results,
                "__ionization_mode": self.__ionization_mode,
                "organized_samples": self.organized_samples,
                "feature_tables": self.feature_tables,
                "empCpds": self.empCpds,
                "log_tranformed_feature_tables": self.log_transformed_feature_tables,
                "cosmetics": self.cosmetics,
                "used_cosmetics": self.used_cosmetics,
                "extracted": self.extracted,
                "subdirectories": self.subdirectories
            }
            try:
                JSON_repr['acquisitions'] = [acquisition.JSON_repr for acquisition in self.acquisitions]
                test = json.dumps(JSON_repr)
                json.dump(JSON_repr, save_filehandle, indent=4)
                shutil.move(save_path, save_path.replace(".tmp",''))
            except:
                pass

    def add_acquisition(self, acquisition, mode="link"):
        """
        This method adds an acquisition to the list of acquisitions in the experiment, ensures there are no duplicates
        and then links or copies the acquisition, currently only as a .raw file, to the experiment directory

        Args:
            acquisition (Acquisition object): the object representing an acquistiion
            override (bool, optional): if True, duplicate names or source paths are allowed. Defaults to False.
            mode (str, optional): if 'link', use symlink, if 'copy' copy the .raw file. Defaults to "link".
        """        
        addition_modes = {
            "copy": shutil.copy,
            "link": os.link,
        }
        if os.path.exists(acquisition.source_filepath):
            if acquisition.source_filepath.endswith(".mzML"):
                target_path = os.path.join(self.converted_subdirectory, os.path.basename(acquisition.source_filepath))
                acquisition.mzml_filepath = target_path
                acquisition.raw_filepath = None
            elif acquisition.source_filepath.endswith(".raw"):
                target_path = os.path.join(self.raw_subdirectory, os.path.basename(acquisition.source_filepath))
                acquisition.mzml_filepath = None
                acquisition.raw_filepath = target_path
            if not os.path.exists(target_path):
                addition_modes[mode](acquisition.source_filepath, target_path)
                self.acquisitions.append(acquisition)
            else:
                print("Target already exists: ", acquisition.name, acquisition.source_filepath)
        else:
            print("File Not Found: ", acquisition.name, acquisition.source_filepath)

    def convert_raw_to_mzML(self, conversion_command):
        """
        Convert all raw files to mzML

        Args:
            mono_path (str): path to mono executable
            exe_path (str): path to converter executable
        """ 
        raw_acquisitions = [acquisition for acquisition in self.acquisitions if acquisition.mzml_filepath is None]
        jobs = []
        output_paths = []
        for acquisition in raw_acquisitions:
            input_filepath = acquisition.raw_filepath
            output_filepath = os.path.join(self.converted_subdirectory, os.path.basename(acquisition.raw_filepath)).replace(".raw", ".mzML")
            output_paths.append(output_filepath)
            new_job = []
            if type(conversion_command) is list:
                for element in conversion_command:
                    if element == "$RAW_PATH":
                        new_job.append(input_filepath)
                    elif element == "$OUT_PATH":
                        new_job.append(output_filepath)
                    else:
                        new_job.append(element)
                jobs.append(new_job)
            elif type(conversion_command is str):
                conversion_command = conversion_command.replace("$RAW_PATH", input_filepath)
                conversion_command = conversion_command.replace("$OUT_PATH", output_filepath) 
        print(jobs)
        workers = mp.Pool(mp.cpu_count())
        for _, raw_acquisition, output_path in zip(workers.map(os.system, [' '.join(x) for x in jobs]), raw_acquisitions, output_paths):
            print(output_path)
            raw_acquisition.mzml_filepath = output_path
            

    def filter_samples(self, filter, return_field=None):
        """
        Find the set of acquisitions that pass the provided filter and return either the acquisition object or the 
        specified field of each passing sample

        Args:
            filter (dict): a filter dict compatable with acquisition.filter
            return_field (str, optional): if defined, return that field from the acquisition, else the object. Defaults to "name".

        Returns:
            list: list of passing acquisitions or fields from passing acquisitions
        """        
        if return_field:
            return [getattr(acquisition, return_field) for acquisition in self.acquisitions if acquisition.filter(filter)]
        else:
            return [acquisition for acquisition in self.acquisitions if acquisition.filter(filter)]

    def construct_experiment_from_CSV(experiment_directory, CSV_filepath, ionization_mode, filter=None, name_field='Name', path_field='Filepath', exp_config=None):
        """
        For a given sequence file, create the experiment object, including the addition of acquisitions

        Args:
            experiment_directory (str): path to the directory with experiment data and intermediates
            CSV_filepath (str): filepath to the sequence CSV
            ionization_mode (str): 'pos' or 'neg' specifies the ionization mode for the experiment
            filter (str, optional): json string representing a dictionary to filter acquistions on. Defaults to None.
            name_field (str, optional): the field to become the acquisition name. Defaults to 'Name'.
            path_field (str, optional): the field that specifies the acquisition filepath. Defaults to 'Filepath'.

        Returns:
            experiment: the experiment object
        """        
        if not (os.path.exists(experiment_directory) or os.path.exists(os.path.join(experiment_directory, "experiment.json"))):
            filter = {} if filter is None else filter
            experiment = Experiment('', experiment_directory, sub_dirs=exp_config["experiment_subdirectories"])
            with open(CSV_filepath, encoding='utf-8-sig') as CSV_fh:
                for acquisition_info in csv.DictReader(CSV_fh):
                    acquisition_info = {k.strip(): v.strip() for k,v in acquisition_info.items()}
                    if name_field not in acquisition_info or path_field not in acquisition_info:
                        raise Exception()
                    acquisition = Acquisition.Acquisition(acquisition_info[name_field], acquisition_info[path_field], acquisition_info)
                    acquisition.experiment = experiment
                    if acquisition.filter(filter):
                        experiment.add_acquisition(acquisition)
            if experiment.acquisitions:
                if exp_config:
                    Experiment.subdirectories = exp_config["experiment_subdirectories"]
                experiment.__ionization_mode = ionization_mode
                experiment.save()
                return experiment
            else:
                import shutil
                shutil.rmtree(experiment.experiment_directory)
                print("Unable to create empty experiment")
                exit()
        else:
            print("Experiment already exists!")
            exit()
    
    def determine_ionization_mode(self, num_files_to_check=5):
        tested = []
        ion_modes = set()
        while len(tested) < num_files_to_check:
            acquisition = random.sample([x for x in self.acquisitions if x not in tested], 1)[0]
            ionization_mode = acquisition.ionization_mode
            tested.append(acquisition)
            ion_modes.add(ionization_mode)
        if len(ion_modes) == 1:
            self.__ionization_mode = list(ion_modes)[0]
        else:
            raise Exception()
        return self.__ionization_mode

    def summarize(self):
        """
        Print the list of empCpds and feature tables in the experiment to the console
        """        
        print("empCpds:")
        for moniker, path in self.empCpds.items():
            print("\t", moniker, " - ", path)
        print("feature tables:")
        for moniker, path in self.feature_tables.items():
            print("\t", moniker, " - ", path)

    def generate_cosmetic_map(self, field=None, cos_type='color', seed=None):
        if seed:
            random.seed(seed)
        if cos_type in self.cosmetics and field in self.cosmetics[cos_type]:
            return self.cosmetics[cos_type][field]
        else:
            if cos_type == 'color':
                banned_colors = self.parameters["banned_colors"]
                allowed_colors = [x for x in mcolors.CSS4_COLORS if x not in banned_colors]
                allowed_cosmetics = [x for x in allowed_colors if x not in self.used_cosmetics]
            elif cos_type == 'marker':
                allowed_markers = self.parameters["allowed_markers"]
                allowed_cosmetics = [x for x in allowed_markers if x not in self.used_cosmetics]
            elif cos_type == 'text':
                pass
            needs_cosmetic = list(set([acquisition.metadata_tags[field] for acquisition in self.acquisitions]))
            cosmetic_map = {t: c for t, c in zip(needs_cosmetic, random.sample(allowed_cosmetics, len(needs_cosmetic)))}   
            self.used_cosmetics.extend(list(cosmetic_map.values()))
            if cos_type not in self.cosmetics:
                self.cosmetics[cos_type] = {}
            self.cosmetics[cos_type][field] = cosmetic_map
            self.save()
            return cosmetic_map

    def batches(self, batch_field):
        batches = {}
        for acquisition in self.acquisitions:
            batch_field_value = acquisition.metadata_tags[batch_field]
            if batch_field_value not in batches:
                batches[batch_field_value] = []
            batches[batch_field_value].append(acquisition.name)
        return batches
    
    def enumerate(self, metadata_field):
        return list(set([x.metadata_tags[metadata_field] for x in self.acquisitions]))
    
    def list_metadata_fields(self):
        for k in self.acquisitions[0].metadata_tags.keys():
            print(k)
            observed = set()
            for acquisition in self.acquisitions:
                v = acquisition.metadata_tags[k]
                if v not in observed:
                    print("\t", v)
                    observed.add(v)
    
    def asari(self, asari_cmd):
        import subprocess
        mapping = {
            '$IONIZATION_MODE': self.ionization_mode,
            '$CONVERTED_SUBDIR': self.converted_subdirectory,
            '$ASARI_SUBDIR': self.asari_subdirectory
        }
        job = []
        if type(asari_cmd) is list:
            for field in asari_cmd:
                job.append(mapping.get(field, field))
        if type(asari_cmd) is str:
            for key, value in mapping.items():
                asari_cmd.replace(key, value)
            job = asari_cmd.split(" ")
        completed_process = subprocess.run(job)
        if completed_process.returncode == 0:
            self.feature_tables['full'] = os.path.join(self.asari_subdirectory, os.listdir(self.asari_subdirectory)[0], "export/full_Feature_table.tsv") 
            self.feature_tables['preferred'] = os.path.join(self.asari_subdirectory, os.listdir(self.asari_subdirectory)[0], "preferred_Feature_table.tsv") 
            self.empCpds['asari'] = os.path.join(self.asari_subdirectory, os.listdir(self.asari_subdirectory)[0], "Annotated_empiricalCompounds.json")
            
